import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import pytplot
from pytplot import get_data
import pyspedas
from pyspedas import time_double
from pyspedas.mms.fpi.mms_get_fpi_dist import mms_get_fpi_dist
from pyspedas.mms.fpi.mms_pad_fpi import mms_pad_fpi


def mms_fpi_ang_ang(time,
                    species='i',
                    probe='1',
                    data_rate='fast',
                    fgm_data_rate='srvy',
                    level='l2',
                    fgm_level='l2',
                    center_measurement=False,
                    energy_range=[10, 30000],
                    xsize=8,
                    ysize=8,
                    cmap='spedas',
                    zrange=[None, None],
                    nocontours=False,
                    save_png=None,
                    save_jpeg=None,
                    save_eps=None,
                    save_svg=None,
                    dpi=300,
                    display=True):
    """
    Creates various plots directly from the FPI distribution functions, including:
         - angle-angle (azimuth vs zenith)
         - angle-energy (azimuth and zenith vs energy)

    Input
    ------
        time: str
            Time to generate the angle-angle plots for

    Parameters
    -----------
        species: str
            'i' for ions, 'e' for electrons

        probe: str
            MMS spacecraft probe # (1-4)

        data_rate: str
            Data rate ('brst' or 'fast')

        level: str
            Data level

        center_measurement: bool
            Flag to center the FPI measurements

        energy_range: list
            Range of energies to include in the figures (default: 10-30000 eV)

        xsize: float
            X-size of the output figure

        ysize: float
            Y-size of the output figure

        cmap: str
            Color map to use (default: 'spedas')

        zrange: list
            Range of the color bar

        display: bool
            Flag to display the figures (with plt.show()); default: True

    """

    trange = [time_double(time)-300, time_double(time)+300]

    fpi_vars = pyspedas.mms.fpi(trange=trange,
                                datatype=['d'+species+'s-moms', 'd'+species+'s-dist'],
                                probe=probe,
                                data_rate=data_rate,
                                level=level,
                                center_measurement=center_measurement)

    fgm_vars = pyspedas.mms.fgm(trange=trange, probe=probe, data_rate=fgm_data_rate, level=fgm_level)
    dist = get_data('mms'+probe+'_d'+species+'s_dist_'+data_rate)

    closest_idx = np.searchsorted(dist.times, time_double(time), side='left')

    data = dist.y[closest_idx, :, :, :]

    if len(dist.v1.shape) > 1:
        phi = dist.v1[closest_idx, :]
    else:
        phi = dist.v1

    theta = dist.v2
    energies = dist.v3[closest_idx, :]
    energy_axis = [np.nanmin(energies), np.nanmax(energies)]

    idx_of_ens = np.argwhere((energies >= energy_range[0]) & (energies <= energy_range[1])).flatten()
    energies = energies[idx_of_ens]
    data_at_ens = data[:, :, idx_of_ens]
    data_summed = np.nansum(data_at_ens, axis=2)

    dists = mms_get_fpi_dist('mms'+probe+'_d'+species+'s_dist_'+data_rate,
                             single_time=time)

    # theta is stored as co-latitude
    theta_colat = dists[0]['theta'][0, 0, :]

    # convert to latitude
    theta_flow_direction = 90.0 - theta_colat
    phi = dists[0]['phi'][0, :, 0]

    spec_options = {}

    pa_dist = mms_pad_fpi(dists,
                      time=dist.times[closest_idx],
                      mag_data='mms'+probe+'_fgm_b_gse_'+fgm_data_rate+'_'+fgm_level+'_bvec')

    # the first figure: azimuth vs. zenith
    fig, axes = plt.subplots()
    fig.set_size_inches(xsize, ysize)

    # add support for the 'spedas' color bar
    if cmap == 'spedas':
        _colors = pytplot.spedas_colorbar
        spd_map = [(np.array([r, g, b])).astype(np.float64)/256 for r, g, b in zip(_colors.r, _colors.g, _colors.b)]
        spec_options['cmap'] = LinearSegmentedColormap.from_list('spedas', spd_map)
    else:
        spec_options['cmap'] = cmap

    # phi can be out of order, so we need to sort prior to sending to pcolormesh
    phi_sorted_idx = np.argsort(phi)
    phi_sorted = phi[phi_sorted_idx]
    data_phi_sorted = data_summed[phi_sorted_idx, :]

    spec_options['norm'] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im = axes.pcolormesh(phi_sorted, theta_flow_direction, data_phi_sorted.T, **spec_options)

    axes.set_xlabel('Az flow angle (deg)')
    axes.set_ylabel('Zenith flow angle (deg)')
    fig.subplots_adjust(left=0.14, right=0.85)
    box = axes.get_position()
    pad, width = 0.02, 0.02
    cax = fig.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar = fig.colorbar(im, cax=cax)
    colorbar.set_label('f ($s^3$/$cm^6$)')

    if not nocontours:
        num_levels = 16
        contour_levels = np.array(180*np.arange(num_levels+1)/num_levels, dtype=int)
        contours = axes.contour(pa_dist['wpol'], pa_dist['waz'], pa_dist['pa_azpol'], contour_levels, linewidths=0.5)
        axes.clabel(contours, contours.levels, inline=True, fontsize=10)

    if save_png is not None and save_png != '':
        plt.savefig(save_png + '_azimuth_vs_zenith.png', dpi=dpi)

    if save_jpeg is not None and save_jpeg != '':
        plt.savefig(save_jpeg + '_azimuth_vs_zenith.jpeg', dpi=dpi)

    if save_eps is not None and save_eps != '':
        plt.savefig(save_eps + '_azimuth_vs_zenith.eps', dpi=dpi)

    if save_svg is not None and save_svg != '':
        plt.savefig(save_svg + '_azimuth_vs_zenith.svg', dpi=dpi)

    # Zenith vs. energy
    fig2, axes2 = plt.subplots()
    fig2.set_size_inches(xsize, ysize)

    theta_en = np.nansum(data_at_ens, axis=0)

    axes2.set_xscale('log')
    axes2.set_xlabel('Energy (eV)')
    axes2.set_ylabel('Zenith flow angle (deg)')
    fig2.subplots_adjust(left=0.14, right=0.85)

    zrange_is_None = zrange[0] is None

    if zrange_is_None:
        zrange[0] = np.nanmin(theta_en[theta_en > 0])

    spec_options['norm'] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im2 = axes2.pcolormesh(energies, theta_flow_direction, theta_en, **spec_options)

    cax2 = fig2.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar2 = fig2.colorbar(im2, cax=cax2)
    colorbar2.set_label('f ($s^3$/$cm^6$)')

    if save_png is not None and save_png != '':
        plt.savefig(save_png + '_zenith_vs_energy.png', dpi=dpi)

    if save_jpeg is not None and save_jpeg != '':
        plt.savefig(save_jpeg + '_zenith_vs_energy.jpeg', dpi=dpi)

    if save_eps is not None and save_eps != '':
        plt.savefig(save_eps + '_zenith_vs_energy.eps', dpi=dpi)

    if save_svg is not None and save_svg != '':
        plt.savefig(save_svg + '_zenith_vs_energy.svg', dpi=dpi)

    # Azimuth vs. energy
    fig3, axes3 = plt.subplots()
    fig3.set_size_inches(xsize, ysize)

    phi_en = np.nansum(data_at_ens[phi_sorted_idx, :, :], axis=1)

    fig3.subplots_adjust(left=0.14, right=0.85)
    axes3.set_xscale('log')
    axes3.set_xlabel('Energy (eV)')
    axes3.set_ylabel('Azimuth flow angle (deg)')

    if zrange_is_None:
        zrange[0] = np.nanmin(phi_en[phi_en > 0])

    spec_options['norm'] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im3 = axes3.pcolormesh(energies, phi_sorted, phi_en, **spec_options)

    cax3 = fig3.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar3 = fig3.colorbar(im3, cax=cax3)
    colorbar3.set_label('f ($s^3$/$cm^6$)')

    if save_png is not None and save_png != '':
        plt.savefig(save_png + '_azimuth_vs_energy.png', dpi=dpi)

    if save_jpeg is not None and save_jpeg != '':
        plt.savefig(save_jpeg + '_azimuth_vs_energy.jpeg', dpi=dpi)

    if save_eps is not None and save_eps != '':
        plt.savefig(save_eps + '_azimuth_vs_energy.eps', dpi=dpi)

    if save_svg is not None and save_svg != '':
        plt.savefig(save_svg + '_azimuth_vs_energy.svg', dpi=dpi)

    # PA vs. energy
    fig4, axes4 = plt.subplots()
    fig4.set_size_inches(xsize, ysize)

    fig4.subplots_adjust(left=0.14, right=0.85)
    axes4.set_yscale('log')
    axes4.set_xlabel('Pitch angle (deg)')
    axes4.set_ylabel('Energy (eV)')

    spec_options['norm'] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im4 = axes4.pcolormesh(pa_dist['pa'], pa_dist['egy'], pa_dist['data'], **spec_options)

    cax4 = fig4.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar4 = fig4.colorbar(im4, cax=cax4)
    colorbar4.set_label('f ($s^3$/$cm^6$)')

    if save_png is not None and save_png != '':
        plt.savefig(save_png + '_pa_vs_energy.png', dpi=dpi)

    if save_jpeg is not None and save_jpeg != '':
        plt.savefig(save_jpeg + '_pa_vs_energy.jpeg', dpi=dpi)

    if save_eps is not None and save_eps != '':
        plt.savefig(save_eps + '_pa_vs_energy.eps', dpi=dpi)

    if save_svg is not None and save_svg != '':
        plt.savefig(save_svg + '_pa_vs_energy.svg', dpi=dpi)

    if display:
        plt.show()



import numpy as np
import scipy.interpolate
import scipy.spatial


def slice2d_2di(datapoints, xyz, resolution, thetarange=None, zdirrange=None):
    """
    Produces slice by interpolating projected data. Based on spd_slice2d_2di in IDL
    """

    # cut by theta value
    if thetarange is not None:
        thetarange = [np.nanmin(thetarange), np.nanmax(thetarange)]

        r = np.sqrt(xyz[:, 0]**2 + xyz[:, 1]**2 + xyz[:, 2]**2)
        eachangle = np.arcsin(xyz[:, 2]/r)*180.0/np.pi

        index = np.argwhere((eachangle <= thetarange[1]) & (eachangle >= thetarange[0])).flatten()

        if len(index) != 0:
            if len(index) != len(r):
                xyz = xyz[index, :]
                datapoints = datapoints[index]

    # cut by z-axis value
    if zdirrange is not None:
        zdirrange = [np.nanmin(zdirrange), np.nanmax(zdirrange)]

        index = np.argwhere((xyz[:, 2] >= zdirrange[0]) & (xyz[:, 2] <= zdirrange[1])).flatten()

        if len(index) != 0:
            if len(index) != len(datapoints):
                xyz = xyz[index, :]
                datapoints = datapoints[index]

    x = xyz[:, 0]
    y = xyz[:, 1]

    # average duplicates along the way
    # note: the following algorithm was translated directly
    # from IDL, and apparently came from thm_esa_slice2d
    # originally
    # uni2 = np.unique(x, return_index=True)[1]+1
    uni2 = np.unique(x, return_index=True)[1]
    uni1 = np.insert(uni2[0:len(uni2)-1] + 1, 0, 0)

    kk = 0
    for i in range(0, len(uni2)):
        yi = y[uni1[i]:uni2[i]+1]
        xi = x[uni1[i]:uni2[i]+1]
        datapointsi = datapoints[uni1[i]:uni2[i]+1]

        xi = xi[np.argsort(yi)]
        datapointsi = datapointsi[np.argsort(yi)]
        yi = yi[np.argsort(yi)]

        index2 = np.unique(yi, return_index=True)[1]
        if len(index2) == 1:
            index1 = 0
        else:
            index1 = np.insert(index2[0:len(index2)-1]+1, 0, 0)

        for j in range(0, len(index2)):
            if isinstance(index1, int):
                y[kk] = yi[index1]
                x[kk] = xi[index1]
                if index1 == index2[j]:
                    datapoints[kk] = datapointsi[index1]
                else:
                    datapoints[kk] = np.nanmean(datapointsi[index1:index2[j]+1])
            else:
                y[kk] = yi[index1[j]]
                x[kk] = xi[index1[j]]
                if index1[j] == index2[j]:
                    datapoints[kk] = datapointsi[index1[j]]
                else:
                    datapoints[kk] = np.nanmean(datapointsi[index1[j]:index2[j]+1])
            kk = kk + 1

    y = y[0:kk]
    x = x[0:kk]
    datapoints = datapoints[0:kk]

    xmax = np.nanmax(np.abs(np.array([y, x])))
    xrange = [-1*xmax, xmax]
    xgrid = np.linspace(xrange[0], xrange[1], num=resolution)
    ygrid = np.linspace(xrange[0], xrange[1], num=resolution)
    out = np.zeros((resolution, resolution))
    grid_interpolator = griddata_tri(datapoints, x, y)
    for xi, xp in enumerate(xgrid):
        for yi, yp in enumerate(ygrid):
            out[xi, yi] = grid_interpolator((xp, yp))

    return {'data': out, 'xgrid': xgrid, 'ygrid': ygrid}


def griddata_tri(data, x, y):
    cart_temp = np.array([x, y])
    points = np.stack(cart_temp).T
    delaunay = scipy.spatial.Delaunay(points)
    return scipy.interpolate.LinearNDInterpolator(delaunay, data)



import pytplot
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from .slice2d_getinfo import slice2d_getinfo


def plot(the_slice,
         xrange=None,
         yrange=None,
         zrange=None,
         colormap=None,
         olines=8,
         contours=False,
         plotsize=10,
         save_png=None,
         save_jpeg=None,
         save_svg=None,
         save_pdf=None,
         save_eps=None,
         dpi=None,
         display=True):
    """
    Creates plots of 2D particle slices
    """
    spec_options = {}

    if zrange is None:
        zrange = the_slice.get('zrange')
        # padding, # in thm_ui_slice2d should match this
        zrange = [zrange[0]*0.999, zrange[1]*1.001]

    if zrange is None:
        spec_options['norm'] = mpl.colors.LogNorm()
    else:
        spec_options['norm'] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    style = pytplot.tplot_opt_glob.get('style')

    if style is None:
        if colormap is None:
            colormap = 'spedas'
    else:
        plt.style.use(style)

    if colormap == 'spedas':
        _colors = pytplot.spedas_colorbar
        spd_map = [(np.array([r, g, b])).astype(np.float64) / 256 for r, g, b in zip(_colors.r, _colors.g, _colors.b)]
        cmap = LinearSegmentedColormap.from_list('spedas', spd_map)
    else:
        cmap = colormap

    spec_options['cmap'] = cmap

    char_size = pytplot.tplot_opt_glob.get('charsize')
    if char_size is None:
        char_size = 12

    fig, axes = plt.subplots()
    fig.set_size_inches(plotsize, plotsize)

    if xrange is not None:
        axes.set_xlim(xrange)

    if yrange is not None:
        axes.set_ylim(yrange)

    axis_font_size = pytplot.tplot_opt_glob.get('axis_font_size')

    if axis_font_size is not None:
        axes.tick_params(axis='x', labelsize=axis_font_size)
        axes.tick_params(axis='y', labelsize=axis_font_size)

    info = slice2d_getinfo(the_slice)

    axes.set_title(info['title'], fontsize=char_size)
    axes.set_ylabel(info['ytitle'], fontsize=char_size)
    axes.set_xlabel(info['xtitle'], fontsize=char_size)

    fig.subplots_adjust(left=0.14, right=0.86, top=0.86, bottom=0.14)

    box = axes.get_position()
    pad, width = 0.02, 0.01
    cax = fig.add_axes([box.xmax + pad, box.ymin, width, box.height])

    if axis_font_size is not None:
        cax.tick_params(labelsize=axis_font_size)

    im = axes.pcolormesh(the_slice['xgrid'], the_slice['ygrid'], the_slice['data'].T, **spec_options)

    colorbar = fig.colorbar(im, cax=cax)

    colorbar.set_label(info['ztitle'], fontsize=char_size)

    # draw lines at the origin
    axes.axvline(x=0, linestyle=(0, (5, 10)), color='black')
    axes.axhline(y=0, linestyle=(0, (5, 10)), color='black')

    interp_type = the_slice.get('interpolation')

    if interp_type != 'geometric' or contours:
        levels = 10.**(np.arange(olines)/float(olines)*(np.log10(zrange[1])-np.log10(zrange[0]))+np.log10(zrange[0]))
        contours = axes.contour(the_slice['xgrid'], the_slice['ygrid'], the_slice['data'].T, levels, linewidths=0.5)

    if save_png is not None and save_png != '':
        plt.savefig(save_png + '.png', dpi=dpi)

    if save_jpeg is not None and save_jpeg != '':
        plt.savefig(save_jpeg + '.jpeg', dpi=dpi)

    if save_eps is not None and save_eps != '':
        plt.savefig(save_eps + '.eps', dpi=dpi)

    if save_svg is not None and save_svg != '':
        plt.savefig(save_svg + '.svg', dpi=dpi)

    if save_pdf is not None and save_pdf != '':
        plt.savefig(save_pdf + '.pdf', dpi=dpi)

    if display:
        plt.show()



import logging
import pyspedas
from pyspedas import time_double
from pyspedas.mms.fpi.mms_get_fpi_dist import mms_get_fpi_dist
from pyspedas.mms.hpca.mms_get_hpca_dist import mms_get_hpca_dist
from pyspedas.particles.spd_slice2d.slice2d import slice2d
from pyspedas.particles.spd_slice2d.slice2d_plot import plot


def mms_part_slice2d(trange=None,
                     time=None,
                     samples=None,
                     window=None,
                     center_time=False,
                     erange=None,
                     thetarange=None,
                     zdirrange=None,
                     average_angle=None,
                     sum_angle=None,
                     energy=False,
                     log=False,
                     probe='1',
                     instrument='fpi',
                     data_rate=None,
                     level='l2',
                     spdf=False,
                     mag_data_rate=None,
                     species=None,
                     rotation='xy',
                     custom_rotation=None,
                     subtract_bulk=False,
                     xrange=None,
                     yrange=None,
                     zrange=None,
                     resolution=None,
                     interpolation='geometric',
                     contours=False,
                     smooth=None,
                     save_jpeg=None,
                     save_png=None,
                     save_svg=None,
                     save_pdf=None,
                     save_eps=None,
                     plotsize=10,
                     dpi=None,
                     return_slice=False,
                     cmap=None,
                     display=True):
    """
    This routine creates 2D slices of 3D distribution function data from the FPI and HPCA instruments.
    This is essentially a wrapper around slice2d and slice2d_plot, that loads the data,
    any required support data, calculates the slice with slice2d and plots it with slice2d_plot.

    Parameters
    -----------
        probe: int or str
            MMS spacecraft probe # (1, 2, 3 or 4)

        instrument: str
            MMS plasma instrument (fpi or hpca)

        species: str
            Particle species; depends on the instrument:
            FPI: 'e' for electrons, 'i' for ions
            HPCA: 'hplus' for H+, 'oplus' for O+, 'heplus' for He+, 'heplusplus', for He++

        data_rate: str
            FPI/HPCA data rate [fast (fpi), srvy (hpca), or brst (both fpi and hpca)]

        mag_data_rate: str
            FGM data rate for transformations

        level: str
            Data level (default: l2); lower levels require an SDC username and password

        spdf: bool
            Flag to download the data from SPDF instead of the MMS SDC

        trange: list of str or list of float
            Two-element time range over which data will be averaged (optional; can also use the 'time' keyword)

        time: str or float
            Time at which the slice will be computed (optional; can also use 'trange' above)

        samples: int
            Number of nearest samples to TIME to average.

        window: float
            Length in seconds from TIME over which data will be averaged

        center_time: bool
            Flag denoting that TIME should be midpoint for window instead of beginning

        interpolation: str
            Interpolation method; valid options:

            'geometric': Each point on the plot is given the value of the bin it intersects.
                         This allows bin boundaries to be drawn at high resolutions.
            '2d': Data points within the specified theta or z-axis range are projected onto
                  the slice plane and linearly interpolated onto a regular 2D grid.

        rotation: str
            Aligns the data relative to the magnetic field and/or bulk velocity.
            This is applied after the CUSTOM_ROTATION. (BV and BE are invariant
            between coordinate systems); valid options:

            'BV':  The x axis is parallel to B field; the bulk velocity defines the x-y plane
            'BE':  The x axis is parallel to B field; the B x V(bulk) vector defines the x-y plane
            'xy':  (default) The x axis is along the data's x axis and y is along the data's y axis
            'xz':  The x axis is along the data's x axis and y is along the data's z axis
            'yz':  The x axis is along the data's y axis and y is along the data's z axis
            'xvel':  The x axis is along the data's x axis; the x-y plane is defined by the bulk velocity
            'perp':  The x axis is the bulk velocity projected onto the plane normal to the B field; y is B x V(bulk)
            'perp_xy':  The data's x & y axes are projected onto the plane normal to the B field
            'perp_xz':  The data's x & z axes are projected onto the plane normal to the B field
            'perp_yz':  The data's y & z axes are projected onto the plane normal to the B field

        custom_rotation: str or np.ndarray
            Applies a custom rotation matrix to the data.  Input may be a
            3x3 rotation matrix or a tplot variable containing matrices.
            If the time window covers multiple matrices they will be averaged.
            This is applied before other transformations

        energy: bool
            Flag to plot data against energy (in eV) instead of velocity.

        log: bool
            Flag to apply logarithmic scaling to the radial measure (i.e. energy/velocity).
            (on by default if ENERGY is True)

        erange: list of float
            Two element array specifying the energy range to be used in eV

        thetarange: list of float
            (2D interpolation only) Angle range, in degrees [-90,90], used to calculate slice.
            Default = [-20,20]; will override ZDIRRANGE.

        zdirrange: list of float
            (2D interpolation only) Z-Axis range, in km/s, used to calculate slice.
            Ignored if called with THETARANGE.

        subtract_bulk: bool
            Subtract the bulk velocity vector

        resolution: int
            Integer specifying the resolution along each dimension of the
            slice (defaults:  2D interpolation: 150, geometric: 500)

        smooth: int
            An odd integer >=3 specifying the width of a smoothing window in #
            of points.  Smoothing is applied to the final plot using a gaussian
            convolution. Even entries will be incremented, 0 and 1 are ignored.

        xrange: list of float
            Keyword to limit the range of the slice's x-axis in the plot

        yrange: list of float
            Keyword to limit the range of the slice's y-axis in the plot

        zrange: list of float
            Keyword to limit the range of the slice's color bar

        save_png: str
            Save the plot as a PNG file

        save_eps: str
            Save the plot as an EPS file

        save_pdf: str
            Save the plot as a PDF file

        save_svg: str
            Save the plot as an SVG file

        display: bool
            Flag to allow disabling displaying of the figure

        return_slice: bool
            Flag to return the slice instead of displaying it

    Returns
    --------
        None (but creates a figure), unless return_slice is set to True

    """

    if trange is None:
        if time is None:
            logging.error('Please specify a time or time range over which to compute the slice.')
            return
        trange_data = [time_double(time)-60, time_double(time)+60]
    else:
        trange_data = trange

    if species is None:
        if instrument == 'fpi':
            species = 'e'
        else:
            species = 'hplus'

    if data_rate is None:
        if instrument == 'fpi':
            data_rate = 'fast'
        else:
            data_rate = 'srvy'

    if mag_data_rate is None:
        if data_rate == 'brst':
            mag_data_rate = 'brst'
        else:
            mag_data_rate = 'srvy'

    instrument = instrument.lower()
    data_rate = data_rate.lower()
    level = level.lower()
    probe = str(probe)

    if rotation in ['xy', 'xz', 'yz']:
        load_support = False
    else:
        load_support = True

    if subtract_bulk:
        load_support = True

    if instrument == 'fpi':
        # not supposed to be centered!
        dist_vars = pyspedas.mms.fpi(probe=probe, trange=trange_data, data_rate=data_rate, time_clip=True,
                                     datatype='d' + species + 's-dist', level=level, spdf=spdf)

        dists = mms_get_fpi_dist('mms' + probe + '_d' + species + 's_dist_' + data_rate, probe=probe)
    elif instrument == 'hpca':
        # supposed to be centered!
        dist_vars = pyspedas.mms.hpca(probe=probe, trange=trange_data, data_rate=data_rate, time_clip=True,
                                      datatype='ion', level=level, center_measurement=True, spdf=spdf)

        dists = mms_get_hpca_dist('mms' + probe + '_hpca_' + species + '_phase_space_density', species=species,
                                  probe=probe, data_rate=data_rate)
    else:
        logging.error('Unknown instrument: ' + instrument + '; valid options: fpi, hpca')
        return

    bfield = None
    vbulk = None
    if load_support:
        fgm_support = pyspedas.mms.fgm(probe=probe, trange=trange_data, data_rate=mag_data_rate, time_clip=True, spdf=spdf)
        bfield = 'mms' + probe + '_fgm_b_gse_' + mag_data_rate + '_l2_bvec'

        if instrument == 'fpi':
            fpi_support = pyspedas.mms.fpi(probe=probe, trange=trange_data, data_rate=data_rate, time_clip=True,
                                           datatype='d'+species+'s-moms', level=level, center_measurement=True, spdf=spdf)
            vbulk = 'mms' + probe + '_d' + species + 's_bulkv_gse_' + data_rate
        elif instrument == 'hpca':
            hpca_support = pyspedas.mms.hpca(probe=probe, trange=trange_data, data_rate=data_rate, time_clip=True,
                                             datatype='moments', level=level, center_measurement=True, spdf=spdf)
            vbulk = 'mms' + probe + '_hpca_' + species + '_ion_bulk_velocity'

    the_slice = slice2d(dists, trange=trange, time=time, window=window, samples=samples, center_time=center_time,
                        mag_data=bfield, vel_data=vbulk, rotation=rotation, resolution=resolution, erange=erange,
                        energy=energy, log=log, custom_rotation=custom_rotation, subtract_bulk=subtract_bulk,
                        interpolation=interpolation, thetarange=thetarange, zdirrange=zdirrange, smooth=smooth,
                        average_angle=average_angle, sum_angle=sum_angle)

    if return_slice:
        return the_slice

    plot(the_slice, xrange=xrange, yrange=yrange, zrange=zrange, save_png=save_png, save_svg=save_svg,
         save_pdf=save_pdf, save_eps=save_eps, save_jpeg=save_jpeg, display=display, dpi=dpi, plotsize=plotsize,
         contours=contours, colormap=cmap)



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG

def load(trange=['2018-11-5', '2018-11-6'], 
         probe='1',
         instrument='fgm',
         datatype='up', 
         suffix='', 
         get_support_data=False, 
         varformat=None, 
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the Cluster mission; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.cluster.fgm
        pyspedas.cluster.aspoc
        pyspedas.cluster.cis
        pyspedas.cluster.dwp
        pyspedas.cluster.edi
        pyspedas.cluster.efw
        pyspedas.cluster.peace
        pyspedas.cluster.rapid
        pyspedas.cluster.staff
        pyspedas.cluster.wbd
        pyspedas.cluster.whi

    """

    if not isinstance(probe, list):
        probe = [probe]

    probe = [str(prb) for prb in probe] # these will need to be strings from now on

    out_files = []

    res=24*3600

    if instrument != 'wbd':
        # note: can't use last_version with WBD data due to using wild cards for the times (and not just in the version)
        last_version = True
    else:
        last_version = False

    for prb in probe:
        if instrument == 'fgm':
            if datatype == 'cp':
                pathformat = 'c'+prb+'/cp/%Y/c'+prb+'_cp_fgm_spin_%Y%m%d_v??.cdf'
            else:
                pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'aspoc':
            pathformat = 'c'+prb+'/'+datatype+'/asp/%Y/c'+prb+'_'+datatype+'_asp_%Y%m%d_v??.cdf'
        elif instrument == 'cis':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'dwp':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'edi':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'efw':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'peace':
            pathformat = 'c'+prb+'/'+datatype+'/pea/%Y/c'+prb+'_'+datatype+'_pea_%Y%m%d_v??.cdf'
        elif instrument == 'rapid':
            pathformat = 'c'+prb+'/'+datatype+'/rap/%Y/c'+prb+'_'+datatype+'_rap_%Y%m%d_v??.cdf'
        elif instrument == 'staff':
            pathformat = 'c'+prb+'/'+datatype+'/sta/%Y/c'+prb+'_'+datatype+'_sta_%Y%m%d_v??.cdf'
        elif instrument == 'whi':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'wbd':
            pathformat = 'c'+prb+'/'+instrument+'/%Y/%m/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d%H%M_v??.cdf'
            res = 600.0

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange, res=res)

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, last_version=last_version)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import numpy as np
from pyspedas.analysis.tnormalize import tnormalize


def cart_trans_matrix_make(x, y, z):

    ndim = x.ndim

    if ndim == 2:
        ex = tnormalize(x, return_data=True)
        ey = tnormalize(y, return_data=True)
        ez = tnormalize(z, return_data=True)
        mat_out = np.concatenate([ex, ey, ez], 1).reshape(x.shape[0], 3, 3)

    elif ndim == 1:
        ex = x/np.sqrt((x*x).sum())
        ey = y/np.sqrt((y*y).sum())
        ez = z/np.sqrt((z*z).sum())
        mat_out = np.array([ex, ey, ez]).T

    return mat_out



import re


def mms_file_filter(files, latest_version=False, major_version=False, min_version=None, version=None):
    """
    This function filters a list of MMS data files based on CDF version
    
    Parameters:
        version: str
            Specify a specific CDF version # to return (e.g., cdf_version='4.3.0')

        min_version: str
            Specify a minimum CDF version # to return

        latest_version: bool
            Only return the latest CDF version in the requested time interval

        major_version: bool
            Only return the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

    Returns:
        List of filtered files
    """
    
    if not isinstance(files, list): files = [files]

    # allow the user to specify partial version #s
    if min_version is not None:
        n_declms = len(min_version.split('.'))
        if n_declms == 1:
            min_version = min_version + '.0.0'
        elif n_declms == 2:
            min_version = min_version + '.0'
    elif version is not None:
        n_declms = len(version.split('.'))
        if n_declms == 1:
            version = version + '.0.0'
        elif n_declms == 2:
            version = version + '.0'

    out_files = []
    file_versions = []
    max_major_version = 0
    max_version = 0

    # find all of the version #s, including the max major version and max total version
    for file in files:
        version_found = re.search(r'v([0-9]+)\.([0-9]+)\.([0-9]+)\.cdf$', file)
        if version_found:
            file_version = version_found.groups()
            # vX.Y.Z
            version_X = int(file_version[0])
            version_Y = int(file_version[1])
            version_Z = int(file_version[2])
            file_versions.append((version_X, version_Y, version_Z, file))
            if version_X > max_major_version:
                max_major_version = version_X
            if max_version == 0:
                max_version = (version_X, version_Y, version_Z)
            else:
                if (version_X > max_version[0]) or (version_X == max_version[0] and version_Y > max_version[1]) or (version_X == max_version[0] and version_Y == max_version[1] and version_Z > max_version[2]):
                    max_version = (version_X, version_Y, version_Z)
        else:
            continue

    for file_ver in file_versions:
        if min_version is not None: # MINIMUM file version
            min_version_num = [int(v) for v in min_version.split('.')]
            if (file_ver[0] > min_version_num[0]) or (file_ver[0] == min_version_num[0] and file_ver[1] > min_version_num[1]) or (file_ver[0] == min_version_num[0] and file_ver[1] == min_version_num[1] and file_ver[2] >= min_version_num[2]):
                out_files.append(file_ver[3])
        elif version is not None: # EXACT file version
            exact_version_num = [int(v) for v in version.split('.')]
            if file_ver[0] == exact_version_num[0] and file_ver[1] == exact_version_num[1] and file_ver[2] == exact_version_num[2]:
                out_files.append(file_ver[3])
        elif latest_version is not False: # LATEST (full) version, i.e., latest X.Y.Z
            if file_ver[0] == max_version[0] and file_ver[1] == max_version[1] and file_ver[2] == max_version[2]:
                out_files.append(file_ver[3])
        elif major_version is not False: # LATEST MAJOR version, i.e., latest X in vX.Y.Z
            if file_ver[0] >= max_major_version:
                out_files.append(file_ver[3])
        else:
            out_files.append(file_ver[3])

    return out_files



from copy import deepcopy
import numpy as np


def qtom(qi):
    """
    Transforms quaternions into rotation matrices. Returns the transpose of the rotation matrix.

    This routine is based on the IDL version by Patrick Cruce
    """

    e00 = qi[:, 0] * qi[:, 0]
    e11 = qi[:, 1] * qi[:, 1]
    e22 = qi[:, 2] * qi[:, 2]
    e33 = qi[:, 3] * qi[:, 3]
    e01 = 2 * qi[:, 0] * qi[:, 1]
    e02 = 2 * qi[:, 0] * qi[:, 2]
    e03 = 2 * qi[:, 0] * qi[:, 3]
    e12 = 2 * qi[:, 1] * qi[:, 2]
    e13 = 2 * qi[:, 1] * qi[:, 3]
    e23 = 2 * qi[:, 2] * qi[:, 3]

    mout = np.zeros((len(e00), 3, 3))

    mout[:, 0, 0] = e00 + e11 - e22 - e33
    mout[:, 1, 0] = e12 + e03
    mout[:, 2, 0] = e13 - e02
    mout[:, 0, 1] = e12 - e03
    mout[:, 1, 1] = e00 - e11 + e22 - e33
    mout[:, 2, 1] = e23 + e01
    mout[:, 1, 2] = e23 - e01
    mout[:, 0, 2] = e13 + e02
    mout[:, 2, 2] = e00 - e11 - e22 + e33

    return mout


def qcompose(vec, theta, free=True):
    """
    Compose quaternions from vectors and angles

    This routine is based on the IDL version by Patrick Cruce
    """

    # Constant indicating where sin(theta) is close enough to theta
    epsilon = 1.0e-20

    vi = deepcopy(vec)
    thi = deepcopy(theta)

    # this next block of code moves angles into the range [0,PI)
    if not free:
        thi = thi % np.pi

        idx = np.argwhere(thi < 0)

        if len(idx) > 1:
            thi[idx] += np.pi

    # calculate the vector norm
    norm = np.sqrt(np.nansum(vi*vi, axis=1))

    # decide which quaternions become identity vectors
    idx1 = np.argwhere(norm < epsilon).flatten()
    idx2 = np.argwhere(norm >= epsilon).flatten()

    out_arr = np.zeros((len(norm), 4))

    if len(idx1) > 0:
        out_arr[idx1, 0] = 1.0
        out_arr[idx1, 1:4] = 0.0

    if len(idx2) > 0:
        out_arr[idx2, 0] = np.cos(thi[idx2]/2.0)

        stheta2 = np.sin(thi[idx2]/2.0)

        out_arr[idx2, 1] = (stheta2 * vi[idx2, 0])/norm[idx2]
        out_arr[idx2, 2] = (stheta2 * vi[idx2, 1])/norm[idx2]
        out_arr[idx2, 3] = (stheta2 * vi[idx2, 2])/norm[idx2]

    return out_arr




from pytplot import options
from .load import load

def mag(trange=['2020-06-01', '2020-06-02'],
        datatype='rtn-normal', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Magnetometer (MAG)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
              'rtn-normal': RTN Coordinates in Normal Mode
              'rtn-normal-1-minute': Same as above, but at 1-min resolution
              'rtn-burst': RTN Coordinates in Burst Mode
              'srf-normal': Spacecraft Reference Frame in Normal Mode
              'srf-burst': Spacecraft Reference Frame in Burst Mode 

        level: str
            Data level (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """

    mag_vars = load(instrument='mag', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)
    
    if mag_vars is None or notplot or downloadonly:
        return mag_vars

    if datatype[-5:] == 'burst':
        ytitle = 'SOLO MAG \\ burst'
    elif datatype[-6:] == 'minute':
        ytitle = 'SOLO MAG \\ 1-min'
    else:
        ytitle = 'SOLO MAG'

    if 'B_SRF'+suffix in mag_vars:
        options('B_SRF'+suffix, 'legend_names', ['Bx (SRF)', 'By (SRF)', 'Bz (SRF)'])
        options('B_SRF'+suffix, 'ytitle', ytitle)

    if 'B_RTN'+suffix in mag_vars:
        options('B_RTN'+suffix, 'legend_names', ['Br (RTN)', 'Bt (RTN)', 'Bn (RTN)'])
        options('B_RTN'+suffix, 'ytitle', ytitle)

    return mag_vars

def rpw(trange=['2020-06-15', '2020-06-16'],
        datatype='hfr-surv', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Radio and Plasma Waves (RPW) instrument
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
            Level 1:
                'hfr-surv', 'lfr-surv-asm', 'lfr-surv-bp1', 'lfr-surv-bp2',
                'lfr-surv-cwf', 'lfr-surv-swf', 'tds-surv-hist1d', 'tds-surv-hist2d',
                'tds-surv-mamp', 'tds-surv-rswf', 'tds-surv-stat', 'tds-surv-tswf',
                'tnr-surv'
                (see below for definitions)

            Level 2:
              High Frequency Receiver (HFR):
                'hfr-surv': High Frequency Receiver (HFR) Data in Survey Mode 

              Low Frequency Receiver (LFR):
                'lfr-surv-asm': Averaged Spectral Matrix (ASM) Data in Survey Mode
                'lfr-surv-bp1': Basic Parameters Set 1 (BP1) Data in Survey Mode
                'lfr-surv-bp2': Basic Parameter Set 2 (BP2) Data in Survey Mode 
                'lfr-surv-cwf-b': Continuous Magnetic Waveform (CWF-B) in Survey Mode
                'lfr-surv-cwf-e': Continuous Electric Waveform (CWF-E) in Survey Mode
                'lfr-surv-swf-b': Snapshot Magnetic Waveform (SWF-B) in Survey Mode
                'lfr-surv-swf-e': Snapshot Electric Waveform (SWF-E) in Survey Mode

              Time Domain Sampler (TDS):
                'tds-surv-hist1d': Histogram Set 1 (HIST1D) Data in Survey Mode
                'tds-surv-hist2d': Histogram Set 2 (HIST2D) Data in Survey Mode
                'tds-surv-mamp': Maximum Amplitude (MAMP) Data in Survey Mode
                'tds-surv-rswf-b': Regular Snapshot Waveform (RSWF) Magnetic Field Data in Survey Mode
                'tds-surv-rswf-e': Regular Snapshot Waveform (RSWF) Electric Field Data in Survey Mode
                'tds-surv-stat': Statistical (STAT) Data in Survey Mode 
                'tds-surv-tswf-b': Triggered Snapshot Magnetic Waveform (TSWF-B) in Survey Mode
                'tds-surv-tswf-e': Triggered Snapshot Electric Waveform (TSWF-E) in Survey Mode

            Level 3:                
                'bia-density': Plasma density derived from probe-to-spacecraft potential and electron plasma frequency
                'bia-density-10-seconds': same as above, but median value over 10 s interval
                'bia-efield-10-seconds': Electric field vector in SRF. Median value over 10 s interval
                'bia-scpot-10-seconds': Spacecraft potential with respect to plasma. Median value over 10 s interval
                'tnr-fp': Plasma frequency value derived by the plasma peak tracking (Thermal Noise Receiver (TNR))

        level: str
            Data level (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='rpw', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def swa(trange=['2020-07-22', '2020-07-23'],
        datatype='pas-eflux', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Solar Wind Plasma Analyser (SWA)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    loaded_vars = load(instrument='swa', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)
    
    if loaded_vars is None or notplot or downloadonly:
        return loaded_vars

    if 'eflux'+suffix in loaded_vars:
        options('eflux'+suffix, 'spec', True)
        options('eflux'+suffix, 'ylog', True)
        options('eflux'+suffix, 'zlog', True)

    return loaded_vars

def epd(trange=['2020-06-14', '2020-06-15'],
        datatype='step', 
        mode='hcad', 
        level='l2', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Energetic Particle Detector (EPD)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        mode: str
            EPD mode; Valid options:

        level: str
            Data level (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='epd', trange=trange, level=level, datatype=datatype, mode=mode, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)





import unittest

from pyspedas.mms.spd_mms_load_bss import spd_mms_load_bss
from pyspedas.mms.mms_load_sroi_segments import mms_load_sroi_segments
from pyspedas.mms.mms_load_fast_segments import mms_load_fast_segments
from pyspedas.mms.mms_load_brst_segments import mms_load_brst_segments
from pyspedas.utilities.data_exists import data_exists


class SegmentTestCases(unittest.TestCase):
    def test_sroi(self):
        sroi = mms_load_sroi_segments(trange=['2019-10-01', '2019-11-01'])
        self.assertTrue(data_exists('mms1_bss_sroi'))
        self.assertTrue(len(sroi[0]) == 28)
        self.assertTrue(sroi[0][0] == 1569849345.0)
        self.assertTrue(sroi[1][0] == 1569923029.0)

    def test_brst(self):
        brst = mms_load_brst_segments(trange=['2015-10-16', '2015-10-17'])
        self.assertTrue(len(brst[0]) == 53)
        self.assertTrue(brst[0][0] == 1444975174.0)
        self.assertTrue(brst[1][0] == 1444975244.0)
        self.assertTrue(data_exists('mms_bss_burst'))

    def test_fast(self):
        fast = mms_load_fast_segments(trange=['2015-10-01', '2015-11-01'])
        self.assertTrue(data_exists('mms_bss_fast'))
        self.assertTrue(len(fast[0]) == 35)
        self.assertTrue(fast[0][0] == 1443504404.0)
        self.assertTrue(fast[1][0] == 1443554774.0)

    def test_spd_mms_load_bss(self):
        spd_mms_load_bss(trange=['2015-10-01', '2015-11-01'])
        self.assertTrue(data_exists('mms_bss_fast'))
        self.assertTrue(data_exists('mms_bss_burst'))
        spd_mms_load_bss(trange=['2019-10-01', '2019-11-01'])
        self.assertTrue(data_exists('mms_bss_burst'))
        self.assertTrue(data_exists('mms1_bss_sroi'))

    def test_spd_mms_load_bss_err(self):
        spd_mms_load_bss(trange=['2015-10-01', '2015-11-01'], datatype='brst', include_labels=True)
        self.assertTrue(~data_exists('mms_bss_fast'))


if __name__ == '__main__':
    unittest.main()




from pyspedas.themis.load import load
from pytplot import options

def fgm(trange=['2007-03-23', '2007-03-24'],
        probe='c',
        level='l2',
        suffix='',
        get_support_data=False,
        varformat=None,
        coord=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads Fluxgate magnetometer (FGM) data

    Parameters:
        trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or list of str
            Spacecraft probe letter(s) ('a', 'b', 'c', 'd' and/or 'e')

        level: str or list of str
            Data type; Valid options: 'l1', 'l2'

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        coord: str
             Coordinate system

        varnames: list of str
            List of variable names to load
            (if not specified, all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables.
            If set, this function returns a list of the files downloaded.

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword

    Returns:
        List of tplot variables created.

    """

    varformat = check_args(varformat=varformat, level=level, coord=coord)

    loaded_vars = load(instrument='fgm', trange=trange, level=level,
                suffix=suffix, get_support_data=get_support_data,
                varformat=varformat, varnames=varnames,
                downloadonly=downloadonly, notplot=notplot,
                probe=probe, time_clip=time_clip, no_update=no_update)

    if loaded_vars is None or loaded_vars == []:
        return []

    if not isinstance(level, list):
        level = [level]

    if not isinstance(probe, list):
        probe = [probe]

    fgm_types = ['fgs', 'fgl', 'fgh', 'fge']
    possible_coords = ['gse', 'gsm', 'dsl', 'ssl']

    # set some plot metadata
    for prb in probe:
        for lvl in level:
            if lvl == 'l2':
                for fgm_type in fgm_types:
                    for coord_sys in possible_coords:
                        if 'th'+prb+'_'+fgm_type+'_'+coord_sys+suffix in loaded_vars:
                            options('th'+prb+'_'+fgm_type+'_'+coord_sys+suffix, 'ytitle', 'TH'+prb.upper()+' '+fgm_type.upper())
                            options('th'+prb+'_'+fgm_type+'_'+coord_sys+suffix, 'color', ['b', 'g', 'r'])
                            options('th'+prb+'_'+fgm_type+'_'+coord_sys+suffix, 'legend_names', ['Bx '+coord_sys.upper(), 'By '+coord_sys.upper(), 'Bz '+coord_sys.upper()])
                    if 'th'+prb+'_'+fgm_type+'_btotal'+suffix in loaded_vars:
                        options('th'+prb+'_'+fgm_type+'_btotal'+suffix, 'ytitle', 'TH'+prb.upper()+' '+fgm_type.upper())
                        options('th'+prb+'_'+fgm_type+'_btotal'+suffix, 'legend_names', 'Bmag')

    return loaded_vars


def check_args(varformat=None, level='l2', coord=None):
    """
    Return varformat for fgm data

    Parameters:
        varformat: str, optional
            If is not empty return as is
        level: {'l1', 'l2'}, optional
            Data level
        coord: {'ssl', 'dsl', 'gse', 'gsm'}, optional
             Coordinate system. If set to None all possible values are used. coord is applicable for level='l2'

    Returns: str
        varformat for themis.load
    """

    # DEV COMMENTS:
    # If varformat is set to None then assign default value
    # We construct regular expression that covers the fgm variable format
    # IDL thm_load_xxx function has the following input:
    #   vsnames = 'a b c d e', $
    #   type_sname = 'probe', $
    #   vdatatypes = 'fgl fgh fge', $
    #   vlevels = 'l1 l2', $ - This input is inherited
    #   vL2datatypes = 'fgs fgl fgh fge fgs_btotal fgl_btotal fgh_btotal fge_btotal', $
    #   vL2coord = 'ssl dsl gse gsm none', $
    # Resulted "varformat" for 'l2' and default coodr ('dsl') is:
    #  th?_fgs_dsl th?_fgl_dsl th?_fgh_dsl th?_fge_dsl
    #  th?_fgs_btotal_dsl th?_fgl_btotal_dsl th?_fgh_btotal_dsl th?_fge_btotal_dsl
    #  th?_fgs th?_fgl th?_fgh th?_fge
    #  th?_fgs_btotal th?_fgl_btotal th?_fgh_btotal th?_fge_btotal
    # Resulted "varformat" for 'l1' is:
    #  *fgl* *fgh* *fge*

    import re

    if varformat is None:
        if level == 'l2':
            # Prepare coordinates, we reuse the same regex to check the coord variable
            coord_regexp = '(ssl|dsl|gse|gsm){1}'
            coord_str = coord_regexp

            # If coord is set and it matches the pattern, then include it into varformat
            if coord is not None and re.search(coord_regexp, coord, re.IGNORECASE):
                coord_str = '(' + coord.lower() + '){1}'

            varformat = '^th[a-e]{1}_(fgs|fgl|fgh|fge){1}(_btotal)?(_{1}' + coord_str + ')?$'
        else:  # This case must be 'l1'
            varformat = '^th[a-e]{1}_(fgl|fgh|fge){1}*'  # keep right end open with '*' (wildcard)

    return varformat



import numpy as np


def slice2d_intrange(dists, trange):
    """
    Retrieves the indices of all samples in the specified time range from a particle distribution list.

    Input
    ------
        dists: list of particle distributions

        trange: array of float
            Time range to find distributions in; must be in unix time

    Returns
    -------
        numpy array containing the indices that fall within the specified trange
    """
    out = []

    for idx, dist in enumerate(dists):
        time = dist['start_time'] + (dist['end_time']-dist['start_time'])/2.0
        if trange[0] <= time <= trange[1]:
            out.append(idx)

    return np.array(out)




def erg_units_string(units, units_only=False,simple=False,relativistic=False):
    """
    Return string describing particle data units for labels

    Input:
        units: str
            String describing units

    Returns:
        String containing full units string
    """
    out = ['', 'Unknown', '']
    units = units.lower()

    if simple:
        if units == 'counts':
            out = ['', 'Counts', '']
        elif units == 'rate':
            out = ['Rate (','#/sec',')']
        elif units == 'eflux':
            out = ['Energy Flux (','eV / sec / cm^2 / ster / eV',')']
        elif units == 'flux':
            out = ['Flux (','# / sec / cm^2 / ster / eV',')']
        elif units == 'df':
            if relativistic:
                out = ['PSD (', '(c/MeV/cm)^3', ')']
            else:
                out = ['PSD (', 's^3 / km^6', ')']
        elif units == 'psd':
            if relativistic:
                out = ['PSD (', '(c/MeV/cm)^3', ')']
            else:
                out = ['PSD (', 's^3 / km^6', ')']
        elif units == 'df_cm':
            out = ['PSD (','s^3 / cm^6',')']
        elif units == 'df_km':
            out = ['PSD (','s^3 / km^6',')']
        elif units == 'e2flux':
            out = ['Energy^2 Flux (','eV^2 / sec / cm^2 / ster /eV',')']
        elif units == 'e3flux':
            out = ['Energy^3 Flux (','eV^3 / sec / cm^2 / ster /eV',')']
    else:
        if units == 'counts':
            out = ['','Counts','']
        elif units == 'rate':
            out = ['Rate (','#/sec',')']
        elif units == 'eflux':
            out = ['Energy Flux (','eV/s/'+'$cm^{2}$'+'/str/eV',')']
        elif units == 'flux':
            out = ['Flux (','#/s/'+'$cm^{2}$'+'/str/eV',')']
        elif units == 'df':
            if relativistic:
                out = ['Phase space density (', '(c/MeV/cm)'+'$^{3}$', ')']
            else:
                out = ['Phase space density (', 's'+'$^{3}$'+'/km'+'$^{6}$', ')']
        elif units == 'psd':
            if relativistic:
                out = ['Phase space density (', '(c/MeV/cm)'+'$^{3}$', ')']
            else:
                out = ['Phase space density (', 's'+'$^{3}$'+'/km'+'$^{6}$', ')']
        elif units == 'df_cm':
            out = ['Phase space density (','s'+'$^{3}$'+'/cm'+'$^{6}$',')']
        elif units == 'df_km':
            out = ['Phase space density (','s'+'$^{3}$'+'/km'+'$^{6}$',')']
        elif units == 'e2flux':
            out = ['Energy$^{2}$ Flux (','eV'+'$^{2}$'+'/s/cm'+'$^{2}$'+'/str/eV',')']
        elif units == 'e3flux':
            out = ['Energy$^{3}$ Flux (','eV'+'$^{3}$'+'/s/cm'+'$^{2}$'+'/str/eV',')']

    if units_only:
        return out[1]

    return ''.join(out)


import cdflib
import fnmatch

import numpy as np

from copy import deepcopy
from pytplot import get_data, store_data, options, clip, ylim, zlim
from pyspedas import tnames

from ....satellite.erg.load import load

from .get_sphcntr import get_sphcntr

"""
;Internal routine to get the table of the pixel
;centers from the table of the pixel corners.
"""
def get_pixel_cntr(tbl_array):
    dim_tuple = tbl_array.shape
    rgmax = dim_tuple[0] - 1
    azmax = dim_tuple[1] - 1
    cnttbl = np.zeros(shape=(rgmax, azmax, 2))
    for i in range(rgmax):
        for j in range(azmax):
            axis_0_indices_array = np.repeat(np.array([[i, i + 1, i + 1, i]]).T,4,1).T.reshape(16)
            axis_1_indices_array = np.array([j] * 8 + [j + 1] * 8)
            lonarr=tbl_array[tuple([axis_0_indices_array, axis_1_indices_array, 0])].reshape(4,4)
            latarr=tbl_array[tuple([axis_0_indices_array, axis_1_indices_array, 1])].reshape(4,4)
            pos_array = get_sphcntr(latarr, lonarr)
            cnttbl[i, j, 1] = pos_array[0]
            cnttbl[i, j, 0] = pos_array[1]
    return cnttbl

def sd_fit(
    trange=['2018-10-18/00:00:00','2018-10-18/02:00:00'],
    suffix='',
    site='all',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    uname=None,
    passwd=None,
    time_clip=False,
    ror=True,
    compact=False
):

    valid_sites = [ 'ade', 'adw', 'bks', 'bpk', 'cly', 'cve', 'cvw', 'dce', 'fhe',
    'fhw', 'fir', 'gbr', 'hal', 'han', 'hok', 'hkw', 'inv', 'kap', 'ker', 'kod',
    'ksr', 'mcm', 'pgr', 'pyk', 'rkn', 'san', 'sas', 'sps', 'sto', 'sye',
    'sys', 'tig', 'unw', 'wal', 'zho', 'lyr' ]



    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(' ')
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if 'all' in site_code:
        site_code = valid_sites
    site_code = list(set(site_code).intersection(valid_sites))

    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    
    for site_input in site_code:

        prefix = 'sd_' + site_input + '_'
        file_res = 3600. * 24.
        pathformat = 'ground/radar/sd/fitacf/'+site_input\
                        +'/%Y/sd_fitacf_l2_'+site_input+'_%Y%m%d*.cdf'

        loaded_data_temp = load(pathformat=pathformat, file_res=file_res, trange=trange, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                    varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)
        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                if isinstance(loaded_data_temp, list):
                    if downloadonly:
                        cdf_file = cdflib.CDF(loaded_data_temp[-1])
                        gatt = cdf_file.globalattsget()
                    else:
                        gatt = get_data(loaded_data_temp[-1], metadata=True)['CDF']['GATT']
                elif isinstance(loaded_data_temp, dict):
                    gatt = loaded_data_temp[list(loaded_data_temp.keys())[-1]]['CDF']['GATT']
                print('############## RULES OF THE ROAD ################')
                print(gatt["Rules_of_use"])
                print('############## RULES OF THE ROAD ################')
            except:
                print('printing PI info and rules of the road was failed')


        if (not downloadonly) and (not notplot) and (len(loaded_data_temp) > 0):

            t_plot_name_list = tnames([prefix + 'pwr*', prefix + 'spec*', prefix + 'vlos*'])
            t_plot_name_list = list(set(t_plot_name_list).intersection(loaded_data))
            for t_plot_name in t_plot_name_list:
                clip(t_plot_name, -9000, 9000)
            
            t_plot_name_list = tnames([prefix + 'elev*'])
            t_plot_name_list = list(set(t_plot_name_list).intersection(loaded_data))
            if len(t_plot_name_list) > 5:
                for t_plot_name in t_plot_name_list:
                    clip(t_plot_name, -9000, 9000)

            azim_no_name_list = list(set(tnames(prefix + '*azim_no_?' + suffix)).intersection(loaded_data))
            number_string_list = []
            for azim_no_name in azim_no_name_list:
                number_string_list.append(azim_no_name.split('_')[4][0])

            for number_string in number_string_list:

                site_input_upper = site_input.upper()
                #  ;Set labels for some tplot variables
                options(prefix + 'pwr_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'pwr_' + number_string + suffix, 'ztitle', 'Backscatter power [dB]')
                options(prefix + 'pwr_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'pwr_err_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'pwr_err_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'pwr_err_' + number_string + suffix, 'ztitle', 'power err [dB]')
                options(prefix + 'spec_width_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'spec_width_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'spec_width_' + number_string + suffix, 'ztitle', 'Spec. width [m/s]')
                options(prefix + 'spec_width_err_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'spec_width_err_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'spec_width_err_' + number_string + suffix, 'ztitle', 'Spec. width err [m/s]')

                if not prefix+'vlos_' + number_string + suffix in loaded_data:
                    vlos_notplot_dictionary = load(pathformat=pathformat, file_res=file_res, trange=trange, prefix=prefix,suffix=suffix, get_support_data=get_support_data,
                            varformat='vlos_' + number_string, downloadonly=downloadonly, notplot=True, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)
                    vlos_tplot_name = prefix+'vlos_' + number_string + suffix
                    if len(vlos_notplot_dictionary) > 0:
                        store_data(vlos_tplot_name,
                                data={'x':vlos_notplot_dictionary[vlos_tplot_name]['x'],
                                    'y':vlos_notplot_dictionary[vlos_tplot_name]['y'],
                                    'v1':vlos_notplot_dictionary[vlos_tplot_name]['v'],
                                    'v2':np.arange(vlos_notplot_dictionary[vlos_tplot_name]['y'].shape[2])},
                                attr_dict={'CDF':vlos_notplot_dictionary[vlos_tplot_name]['CDF']})

                        clip(vlos_tplot_name, -9000, 9000)
                        options(vlos_tplot_name, 'spec', 1)
                        loaded_data.append(vlos_tplot_name)
                options(prefix + 'vlos_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'vlos_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'vlos_' + number_string + suffix, 'ztitle', 'Doppler velocity [m/s]')

                options(prefix + 'vlos_err_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'vlos_err_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'vlos_err_' + number_string + suffix, 'ztitle', 'Vlos err [m/s]')
                if prefix + 'elev_angle_' + number_string + suffix in loaded_data:  # need to get_support_data=True
                    options(prefix + 'elev_angle_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                    options(prefix + 'elev_angle_' + number_string + suffix, 'ysubtitle', '[range gate]')
                    options(prefix + 'elev_angle_' + number_string + suffix, 'ztitle', 'Elev. angle [deg]')
                options(prefix + 'echo_flag_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'echo_flag_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'echo_flag_' + number_string + suffix, 'ztitle', '1: iono. echo')
                options(prefix + 'quality_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'quality_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'quality_' + number_string + suffix, 'ztitle', 'quality')
                options(prefix + 'quality_flag_' + number_string + suffix, 'ytitle', site_input_upper+'\nall beams')
                options(prefix + 'quality_flag_' + number_string + suffix, 'ysubtitle', '[range gate]')
                options(prefix + 'quality_flag_' + number_string + suffix, 'ztitle', 'quality flg')

                #;Split vlos_? tplot variable into 3 components
                get_data_vlos = get_data(prefix + 'vlos_' + number_string + suffix)
                if get_data_vlos is not None:
                    if get_data_vlos[1].ndim >= 3:
                        get_metadata_vlos = get_data(prefix + 'vlos_' + number_string + suffix, metadata=True)
                        store_data(prefix + 'vnorth_' + number_string + suffix,
                                    data={'x':get_data_vlos[0],
                                        'y':get_data_vlos[1][:, :, 0],
                                        'v':get_data_vlos[2]},
                                    attr_dict=get_metadata_vlos)
                        options(prefix + 'vnorth_' + number_string + suffix, 'ztitle', 'LOS V Northward [m/s]')
                        loaded_data.append(prefix + 'vnorth_' + number_string + suffix)
                        if get_data_vlos[1].shape[2] >= 1:
                            store_data(prefix + 'veast_' + number_string + suffix,
                                        data={'x':get_data_vlos[0],
                                            'y':get_data_vlos[1][:, :, 1],
                                            'v':get_data_vlos[2]},
                                        attr_dict=get_metadata_vlos)
                            options(prefix + 'veast_' + number_string + suffix, 'ztitle', 'LOS V Eastward [m/s]')
                            loaded_data.append(prefix + 'veast_' + number_string + suffix)
                            if get_data_vlos[1].shape[2] >= 2:
                                store_data(prefix + 'vlos_' + number_string + suffix,
                                            data={'x':get_data_vlos[0],
                                                'y':get_data_vlos[1][:, :, 2],
                                                'v':get_data_vlos[2]},
                                            attr_dict=get_metadata_vlos)
                                options(prefix + 'vlos_' + number_string + suffix, 'ztitle', 'LOS Doppler vel. [m/s]')

                    #;Combine iono. echo and ground echo for vlos
                    v_var_names = ['vlos_','vnorth_','veast_']
                    flag_data = get_data(prefix + 'echo_flag_' + number_string + suffix)
                    for v_var in v_var_names:
                        v_var_data = get_data(prefix + v_var + number_string + suffix)
                        if v_var_data is not None:
                            v_var_metadata = get_data(prefix + v_var + number_string + suffix, metadata=True)
                            g_data_y = np.where(flag_data[1] == 1., np.nan, v_var_data[1])
                            v_var_data_y = np.where(flag_data[1] != 1., np.nan, v_var_data[1])
                            max_rg = np.nanmax(v_var_data[2]) + 1
                            store_data(prefix + v_var + 'iscat_' + number_string + suffix,
                                    data={'x':v_var_data[0],
                                        'y':v_var_data_y,
                                        'v':v_var_data[2]},
                                    attr_dict=v_var_metadata)
                            options(prefix + v_var + 'iscat_' + number_string + suffix, 'ytitle', ' ')
                            options(prefix + v_var + 'iscat_' + number_string + suffix, 'ysubtitle', ' ')
                            options(prefix + v_var + 'iscat_' + number_string + suffix, 'ztitle', ' ')
                            options(prefix + v_var + 'iscat_' + number_string + suffix, 'spec', 1)
                            loaded_data.append(prefix + v_var + 'iscat_' + number_string + suffix)
                            metadata_for_gscat = deepcopy(v_var_metadata)
                            metadata_for_gscat['plot_options']['extras']['fill_color'] = 5  #options like, 'fill_color:5' in IDL, have not implemented.
                            store_data(prefix + v_var + 'gscat_' + number_string + suffix,
                                    data={'x':v_var_data[0],
                                        'y':g_data_y,
                                        'v':v_var_data[2]},
                                    attr_dict=metadata_for_gscat)
                            options(prefix + v_var + 'gscat_' + number_string + suffix, 'ytitle', ' ')
                            options(prefix + v_var + 'gscat_' + number_string + suffix, 'ysubtitle', ' ')
                            options(prefix + v_var + 'gscat_' + number_string + suffix, 'ztitle', ' ')
                            options(prefix + v_var + 'gscat_' + number_string + suffix, 'spec', 1)
                            loaded_data.append(prefix + v_var + 'gscat_' + number_string + suffix)
                            store_data(prefix + v_var + 'bothscat_' + number_string + suffix,
                                    data=[prefix + v_var + 'iscat_' + number_string + suffix,
                                        prefix + v_var + 'gscat_' + number_string + suffix])
                            options(prefix + v_var + 'bothscat_' + number_string + suffix, 'yrange', [0,max_rg])
                            loaded_data.append(prefix + v_var + 'bothscat_' + number_string + suffix)
                            """
                            Currently, '*iscat_*' and '*bothscat_*' are almost same plot outputs.
                            Because, options like, 'fill_color:5' of IDL for '*gscat_*' have not implemented.
                            """

                #;Set the z range explicitly for some tplot variables
                zlim(prefix + 'pwr_' + number_string + suffix, 0., 30.)
                zlim(prefix + 'pwr_err_' + number_string + suffix, 0., 30.)
                zlim(prefix + 'spec_width_' + number_string + suffix, 0., 200.)
                zlim(prefix + 'spec_width_err_' + number_string + suffix, 0., 300.)
                
                # zlim for '*vlos_*scat_*'
                t_names_raw = tnames(prefix + 'vlos_*scat_' + number_string + suffix)
                t_names_remove_space = [t_name.split(' ')[0] for t_name in t_names_raw]
                t_plot_name_list = list(set(t_names_remove_space).intersection(loaded_data))
                for t_plot_name in t_plot_name_list:
                    zlim(t_plot_name, -400., 400.)
                
                # zlim for '*vnorth_*scat_*'
                t_names_raw = tnames(prefix + 'vnorth_*scat_' + number_string + suffix)
                t_names_remove_space = [t_name.split(' ')[0] for t_name in t_names_raw]
                t_plot_name_list = list(set(t_names_remove_space).intersection(loaded_data))
                for t_plot_name in t_plot_name_list:
                    zlim(t_plot_name, -400., 400.)
                
                # zlim for '*veast_*scat_*'
                t_names_raw = tnames(prefix + 'veast_*scat_' + number_string + suffix)
                t_names_remove_space = [t_name.split(' ')[0] for t_name in t_names_raw]
                t_plot_name_list = list(set(t_names_remove_space).intersection(loaded_data))
                for t_plot_name in t_plot_name_list:
                    zlim(t_plot_name, -400., 400.)
                
                zlim(prefix + 'vlos_err_' + number_string + suffix, 0., 300.)

                # ;Fill values --> NaN
                get_data_vars_pwr = get_data(prefix + 'pwr_' + number_string + suffix)
                if get_data_vars_pwr is not None:
                    pwr_y = deepcopy(get_data_vars_pwr[1])
                    indices_array_tuple = np.where(np.isfinite(pwr_y) == False)
                    var_name_list = ['echo_flag_', 'quality_', 'quality_flag_']
                    for var_name in var_name_list:
                        t_plot_name = prefix + var_name + number_string + suffix
                        get_data_vars = get_data(t_plot_name)
                        get_metadata_vars = get_data(t_plot_name, metadata=True)
                        if get_data_vars is not None:
                            val_array = deepcopy(get_data_vars[1].astype(np.float64))
                            val_array[indices_array_tuple] = np.nan
                            store_data(t_plot_name, data={'x':get_data_vars[0],
                                                        'y':val_array,
                                                        'v':get_data_vars[2]},
                                    attr_dict=get_metadata_vars)

                #;Reassign scan numbers for the combined data
                if (prefix + 'scanstartflag_' + number_string + suffix in loaded_data)\
                   and (prefix + 'scanno_' + number_string + suffix in loaded_data):  # need to get_support_data=True
                    t_plot_name = prefix + 'scanstartflag_' + number_string + suffix
                    scanstartflag_data = get_data(t_plot_name)
                    if scanstartflag_data is not None:
                        scflg = abs(scanstartflag_data[1])
                        try:
                            scno = np.full(shape=scflg.shape, fill_value=-1, dtype=np.int64)
                        except:
                            scno = np.full(shape=scflg.shape, fill_value=-1, dtype=np.int32)
                        scno_t = 0
                        scno[0] = scno_t
                        gt_1_indices_array = np.where(scflg > 0)[0]
                        for i in range(gt_1_indices_array.size - 1):
                            scno[gt_1_indices_array[i]:gt_1_indices_array[i+1]] = i
                        scno[gt_1_indices_array[i+1]:] = i + 1
                        t_plot_name = prefix + 'scanno_' + number_string + suffix
                        get_data_var_scanno = get_data(t_plot_name)
                        if get_data_var_scanno is not None:
                            get_metadata_var_scanno = get_data(t_plot_name, metadata=True)
                            store_data(t_plot_name, data={'x':get_data_var_scanno[0],
                                                          'y':scno},
                                                    attr_dict=get_metadata_var_scanno)

            """
            ;Load the position table(s) ;;;;;;;;;;;;;;;;;;
            ;Currently supports SD fitacf CDFs containing up to 4 pos. tables.
            """
            tbllist = ['tbl_0', 'tbl_1' , 'tbl_2', 'tbl_3', 'tbl_4',
            'tbl_5', 'tbl_6' , 'tbl_7', 'tbl_8', 'tbl_9' ]
            timelist = ['time_0','time_1','time_2','time_3', 'time_4',
            'time_5','time_6','time_7','time_8','time_9']

            get_metadata_vars = get_data(loaded_data[-1], metadata=True)
            if get_metadata_vars is not None:
                datfiles = deepcopy(get_metadata_vars['CDF']['FILENAME'])
                position_tbl_dictionary = {}
                for i in range(10):
                    position_tbl_dictionary[str(i)] = {
                        'time_input':[],
                        'tbl_input':[],
                        'cnttbl_input':[]
                    }
                if len(datfiles) > 0:
                    for file_name in datfiles:
                        cdf_file = cdflib.CDF(file_name)
                        cdf_info = cdf_file.cdf_info()
                        all_cdf_variables = cdf_info['rVariables'] + cdf_info['zVariables']
                        timevn = fnmatch.filter(all_cdf_variables, 'Epoch_?')
                        ptblvn = fnmatch.filter(all_cdf_variables, 'position_tbl_?')
                        timevn.sort()
                        ptblvn.sort()
                        for j in range(len(ptblvn)):
                            tv_name = timevn[j]
                            stblno = tv_name.split('_')[-1]
                            pv_name = ptblvn[j]
                            time_array = cdf_file.varget(tv_name)
                            tbl_array = cdf_file.varget(pv_name)
                            cnttbl = get_pixel_cntr(tbl_array)
                            position_tbl_dictionary[stblno]['time_input'] += [time_array[0], time_array[-1]]
                            dim_tuple = tbl_array.shape
                            tbl2_array = tbl_array.reshape(1, dim_tuple[0], dim_tuple[1], dim_tuple[2])
                            cnttbl2_array = cnttbl.reshape(1, dim_tuple[0] - 1, dim_tuple[1] - 1, dim_tuple[2])
                            if len(position_tbl_dictionary[stblno]['tbl_input']) == 0:
                                position_tbl_dictionary[stblno]['tbl_input'] = np.concatenate(
                                                                                [tbl2_array,tbl2_array],
                                                                                axis=0)
                            else:
                                position_tbl_dictionary[stblno]['tbl_input'] = np.concatenate([position_tbl_dictionary[stblno]['tbl_input']
                                                                                            ,tbl2_array,tbl2_array],axis=0)
                            if len(position_tbl_dictionary[stblno]['cnttbl_input']) == 0:
                                position_tbl_dictionary[stblno]['cnttbl_input'] = np.concatenate([cnttbl2_array,cnttbl2_array],axis=0)
                            else:
                                position_tbl_dictionary[stblno]['cnttbl_input'] = np.concatenate([position_tbl_dictionary[stblno]['cnttbl_input']
                                                                                            ,cnttbl2_array,cnttbl2_array],axis=0)
    
                    for t_plot_suffix_number in position_tbl_dictionary.keys():
                        if len(position_tbl_dictionary[t_plot_suffix_number]['time_input']) >= 2:
                            input_tplot_time_array= np.array(position_tbl_dictionary[t_plot_suffix_number]['time_input'])\
                                                    / 1000. - 719528. * 24.* 3600.
                            t_plot_name = prefix + 'position_tbl_' + t_plot_suffix_number + suffix
                            store_data(t_plot_name, data={'x':input_tplot_time_array,
                                            'y':position_tbl_dictionary[t_plot_suffix_number]['tbl_input']})
                            loaded_data.append(t_plot_name)
                            t_plot_name = prefix + 'positioncnt_tbl_' + t_plot_suffix_number + suffix
                            store_data(t_plot_name, data={'x':input_tplot_time_array,
                                            'y':position_tbl_dictionary[t_plot_suffix_number]['cnttbl_input']})
                            loaded_data.append(t_plot_name)


    if compact:  #;Leave only minimal set of the variables if compact=True.
        search_var_list = ['*cpid*', '*channel*', '*int_time*', '*azim_no*', '*pwr_err*', '*spec_width_err*',
                            '*vlos_err*', '*elev_angle*', '*elev_angle_err*', '*phi0*', '*phi0_err*', 
                            '*echo_flag*', '*quality*', '*quality_flag*', '*scanno*', '*scanstartflag*',
                            '*lagfr*', '*smsep*', '*nrang_max*', '*tfreq*', '*noise*', '*num_ave*', '*txpl*',
                            '*vnorth*', '*veast*', '*vlos_bothscat*', '*vlos_iscat*', '*vlos_gscat*',
                            '*vnorth_iscat*', '*vnorth_gscat*', '*vnorth_bothscat*', '*veast_iscat*',
                            '*veast_gscat*', '*veast_bothscat*', '*position_tbl*', '*positioncnt_tbl*']
        delete_tplot_name_list = list(set(tnames(search_var_list)).intersection(loaded_data))
        if len(delete_tplot_name_list) > 0:
            store_data(delete_tplot_name_list, delete=True)
            loaded_data = list(set(loaded_data).difference(delete_tplot_name_list))


    return loaded_data



import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_notplot(self):
        mam_vars = pyspedas.equator_s.mam(notplot=True)
        self.assertTrue('B_xyz_gse%eq_pp_mam' in mam_vars)

    def test_load_mam_data(self):
        mam_vars = pyspedas.equator_s.mam(time_clip=True)
        self.assertTrue(data_exists('B_xyz_gse%eq_pp_mam'))

    def test_load_esa_downloadonly(self):
        esa = pyspedas.equator_s.esa(downloadonly=True)

    def test_load_edi_data(self):
        edi_vars = pyspedas.equator_s.edi()
        self.assertTrue(data_exists('V_ed_xyz_gse%eq_pp_edi'))
        self.assertTrue(data_exists('E_xyz_gse%eq_pp_edi'))

    def test_load_epi_data(self):
        epi_vars = pyspedas.equator_s.epi()
        self.assertTrue(data_exists('J_e_1%eq_pp_epi'))
        self.assertTrue(data_exists('J_e_2%eq_pp_epi'))

    def test_load_ici_data(self):
        ici_vars = pyspedas.equator_s.ici()
        self.assertTrue(data_exists('N_p%eq_pp_ici'))

    def test_load_pcd_data(self):
        pcd_vars = pyspedas.equator_s.pcd()
        self.assertTrue(data_exists('I_ion%eq_pp_pcd'))

    def test_load_sfd_data(self):
        sfd_vars = pyspedas.equator_s.sfd()
        self.assertTrue(data_exists('F_e>0.26%eq_sp_sfd'))


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_unpublished_data(self):
        """
            this test doesn't load any data, since the username/pw is invalid
        """
        # no password
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN', username='hello')
        # invalid password
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN', username='hello', password='world')
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_SC', username='hello', password='world')
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_SC_1min', username='hello', password='world')
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN_1min', username='hello', password='world')
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_RTN_4_Sa_per_Cyc', username='hello', password='world')
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_SC_4_Sa_per_Cyc', username='hello', password='world')
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='sqtn_rfs_V1V2', username='hello', password='world')
        spc = pyspedas.psp.spc(trange=['2018-11-5', '2018-11-5/06:00'], username='hello', password='world')
        spi = pyspedas.psp.spi(trange=['2018-11-5', '2018-11-5/06:00'], username='hello', password='world')

    def test_load_dfb_dbm_dvac(self):
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='dfb_dbm_dvac', level='l2')
        self.assertTrue(data_exists('psp_fld_l2_dfb_dbm_dvac12'))

    def test_load_fld_data(self):
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn', level='l2', time_clip=True)
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN'))
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_mag_RTN', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_004016'))
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_mag_RTN', 0)
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_000'))
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_mag_RTN', [4, 16], keep=True)

    def test_load_fld_1min(self):
        fields_vars = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn_1min', level='l2')
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_mag_RTN_1min', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_1min'))
        self.assertTrue(data_exists('psp_fld_l2_quality_flags'))
        notplot = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn_1min', level='l2', notplot=True)
        self.assertTrue('psp_fld_l2_mag_RTN_1min' in notplot.keys())

    def test_load_fld_rtn_4_per_cyc(self):
        fields = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_rtn_4_per_cycle', level='l2')
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_mag_RTN_4_Sa_per_Cyc', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_RTN_4_Sa_per_Cyc'))
        self.assertTrue(data_exists('psp_fld_l2_quality_flags'))

    def test_load_fld_sc_4_per_cyc(self):
        fields = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='mag_sc_4_per_cycle',
                                     level='l2')
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_mag_SC_4_Sa_per_Cyc', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_mag_SC_4_Sa_per_Cyc'))
        self.assertTrue(data_exists('psp_fld_l2_quality_flags'))

    def test_load_sqtn_rfs_v1v2(self):
        fields = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='sqtn_rfs_v1v2')
        filtered = pyspedas.psp.filter_fields('electron_density', [4, 16])
        self.assertTrue(data_exists('electron_density'))
        self.assertTrue(data_exists('electron_core_temperature'))

    def test_load_dfb_dc_spec(self):
        fields = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='dfb_dc_spec')
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_dfb_dc_spec_dV12hg', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_dfb_dc_spec_dV12hg'))
        self.assertTrue(data_exists('psp_fld_l2_dfb_dc_spec_SCMdlfhg'))

    def test_load_dfb_ac_xspec(self):
        fields = pyspedas.psp.fields(trange=['2018-11-5', '2018-11-5/06:00'], datatype='dfb_ac_xspec')
        filtered = pyspedas.psp.filter_fields('psp_fld_l2_dfb_ac_xspec_power_ch1_SCMdlfhg', [4, 16])
        self.assertTrue(data_exists('psp_fld_l2_dfb_ac_xspec_power_ch1_SCMdlfhg'))
        self.assertTrue(data_exists('psp_fld_l2_dfb_ac_xspec_power_ch1_SCMdlfhg'))

    def test_load_spc_data(self):
        spc_vars = pyspedas.psp.spc(trange=['2018-11-5', '2018-11-6'], datatype='l3i', level='l3')
        self.assertTrue(data_exists('psp_spc_np_fit'))
        self.assertTrue(data_exists('psp_spc_np_fit_uncertainty'))
        self.assertTrue(data_exists('psp_spc_wp_fit'))
        self.assertTrue(data_exists('psp_spc_vp_fit_SC'))
        self.assertTrue(data_exists('psp_spc_vp_fit_RTN'))
        self.assertTrue(data_exists('psp_spc_np1_fit'))

    def test_load_spe_data(self):
        spe_vars = pyspedas.psp.spe(trange=['2018-11-5', '2018-11-6'], datatype='spa_sf1_32e', level='l2')
        self.assertTrue(data_exists('psp_spe_EFLUX'))
        self.assertTrue(data_exists('psp_spe_QUALITY_FLAG'))

    def test_load_spi_data(self):
        spi_vars = pyspedas.psp.spi(trange=['2018-11-5', '2018-11-6'], datatype='spi_sf0a_mom_inst', level='l3')
        self.assertTrue(data_exists('psp_spi_DENS'))
        self.assertTrue(data_exists('psp_spi_VEL'))
        self.assertTrue(data_exists('psp_spi_T_TENSOR'))
        self.assertTrue(data_exists('psp_spi_TEMP'))
        self.assertTrue(data_exists('psp_spi_EFLUX_VS_ENERGY'))
        self.assertTrue(data_exists('psp_spi_EFLUX_VS_THETA'))
        self.assertTrue(data_exists('psp_spi_EFLUX_VS_PHI'))

    def test_load_epihi_data(self):
        epihi_vars = pyspedas.psp.epihi(trange=['2018-11-5', '2018-11-5/06:00'], datatype='let1_rates1h', level='l2')
        self.assertTrue(data_exists('psp_epihi_B_He_Rate'))
        self.assertTrue(data_exists('psp_epihi_R1A_He_BIN'))
        self.assertTrue(data_exists('psp_epihi_R3B_He_BIN'))
        self.assertTrue(data_exists('psp_epihi_R6A_He_BIN'))

    def test_load_epi_data(self):
        epilo_vars = pyspedas.psp.epi()
        self.assertTrue(data_exists('psp_isois_HET_A_Electrons_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_HET_A_H_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_A_H_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_A_Heavy_Rate_TS'))
        self.assertTrue(data_exists('psp_isois_H_CountRate_ChanP_SP'))
        self.assertTrue(data_exists('psp_isois_Electron_CountRate_ChanE'))

    def test_downloadonly(self):
        files = pyspedas.psp.epilo(downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



"""
@Author : Xin Cao, Xiangning Chu, Sep. 2021.
@University of Colorado Boulder
"""
import os
import numpy as np
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip

from .config import CONFIG
import zipfile
import pyspedas
import logging
import shutil
import gzip

def load(trange = None, resolution=10, dtype = None, no_download = False, downloadonly = False, out_type = 'np', save_pickle = False):
    """
    This function loads SECS/EICS data; this function is not meant
    to be called directly; instead, see the wrapper:
        pyspedas.secs.data

    """

    if dtype == 'EICS' or dtype == 'SECS':
        pathformat_prefix = dtype + '/%Y/%m/'
        pathformat_zip = pathformat_prefix + dtype + '%Y%m%d.zip'
        pathformat_gz = pathformat_prefix + dtype + '%Y%m%d.zip.gz' # only 2007!
        pathformat_unzipped = pathformat_prefix + '%d/' + dtype + '%Y%m%d_%H%M%S.dat'

    else:
        raise TypeError("%r are invalid keyword arguments" % dtype)

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat_zip, trange=trange)
    remote_names_gz = dailynames(file_format=pathformat_gz, trange=trange)
    remote_names_gz = [s for s in remote_names_gz if s[-15:-11] == '2007']

    out_files = []
    out_files_zip = []

    files_zip = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'],
                     local_path=CONFIG['local_data_dir'], no_download=no_download)
    files_gz = download(remote_file=remote_names_gz, remote_path=CONFIG['remote_data_dir'],
                         local_path=CONFIG['local_data_dir'], no_download=no_download)
    files_zip = files_zip + files_gz

    if files_zip is not None:
        for rf_zip_zero in files_zip:
            if rf_zip_zero.endswith('.gz'):
                rf_zip = rf_zip_zero[0:-3]
                # unzip .gz file to .zip file
                with gzip.open(rf_zip_zero, 'rb') as f_in:
                    with open(rf_zip, 'wb') as f_out:
                        shutil.copyfileobj(f_in, f_out)
            elif rf_zip_zero.endswith('.zip'):
                rf_zip = rf_zip_zero
            else:
                rf_zip = rf_zip_zero
            out_files_zip.append(rf_zip)
            #print('Start for unzipping process ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
            foldername_unzipped = rf_zip[0:-19] + rf_zip[-8:-6] + '/' + rf_zip[-6:-4]
            #print('foldername_unzipped-------: ', foldername_unzipped)
            ### add??????
            if not os.path.isdir(foldername_unzipped):
                logging.info('Start unzipping: '+ rf_zip + '  ------')
                with zipfile.ZipFile(rf_zip, 'r') as zip_ref:
                    zip_ref.extractall(rf_zip[0:-16])
                if not os.path.isdir(foldername_unzipped):
                    # for the case of unzipping directly without the %d folder made.
                    # make %d folder
                    os.makedirs(foldername_unzipped)
                    # move .dat files
                    sourcepath = rf_zip[0:-16]
                    sourcefiles = os.listdir(sourcepath)
                    destinationpath = foldername_unzipped
                    logging.info('start to move files: --------------')
                    for file in sourcefiles:
                        if rf_zip[-16:-4] in file and file.endswith('.dat'):
                            shutil.move(os.path.join(sourcepath, file), os.path.join(destinationpath, file))

            else:
                logging.info('Unzipped folder: ' + foldername_unzipped + ' existed, skip unzipping  ------')

    if files_zip is not None:
        for file in files_zip:
            out_files.append(file)
    out_files = sorted(out_files)

    if out_files_zip is not None:
        out_files_zip = list(set(out_files_zip))
        out_files_zip = sorted(out_files_zip)


    if downloadonly:
        return out_files_zip #out_files

    remote_names_unzipped = dailynames(file_format=pathformat_unzipped, trange=trange, res=resolution)
    """
    files_unzipped = download(remote_file=remote_names_unzipped, remote_path=CONFIG['remote_data_dir'],
                         local_path=CONFIG['local_data_dir'], no_download=True)
    """
    remote_names_unzipped_existed = [rnud for rnud in remote_names_unzipped for ofz in out_files_zip if ofz[-16:-4] in rnud]
    remote_names_unzipped = remote_names_unzipped_existed
    out_files_unzipped = [CONFIG['local_data_dir'] + rf_res for rf_res in remote_names_unzipped]
    out_files_unzipped = sorted(out_files_unzipped)

    if out_files_unzipped == []:
        data_vars = []
    else:
        data_vars = pyspedas.secs.read_data_files(out_files=out_files_unzipped, dtype = dtype, out_type = out_type, save_pickle = save_pickle)
        #print('data_vars: ', data_vars, np.shape(data_vars))


    return data_vars #tvars







from pytplot import options
from pyspedas import tnames

def mms_edi_set_metadata(probe, data_rate, level, suffix=''):
    """
    This function updates the metadata for EDI data products
    
    Parameters
    ----------
        probe : str or list of str
            probe or list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate : str or list of str
            instrument data rate for EDI

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

    """
    if not isinstance(probe, list): probe = [probe]
    if not isinstance(data_rate, list): data_rate = [data_rate]
    if not isinstance(level, list): level = [level]

    instrument = 'edi'

    tvars = set(tnames())

    for this_probe in probe:
        for this_dr in data_rate:
            for this_lvl in level:
                if 'mms'+str(this_probe)+'_'+instrument+'_vdrift_dsl_'+this_dr+'_'+this_lvl+suffix in tvars:
                    options('mms'+str(this_probe)+'_'+instrument+'_vdrift_dsl_'+this_dr+'_'+this_lvl+suffix, 'ytitle', 'MMS'+str(this_probe)+' EDI drift velocity')
                    options('mms'+str(this_probe)+'_'+instrument+'_vdrift_dsl_'+this_dr+'_'+this_lvl+suffix, 'legend_names', ['Vx DSL', 'Vy DSL', 'Vz DSL'])
                if 'mms'+str(this_probe)+'_'+instrument+'_vdrift_gse_'+this_dr+'_'+this_lvl+suffix in tvars:
                    options('mms'+str(this_probe)+'_'+instrument+'_vdrift_gse_'+this_dr+'_'+this_lvl+suffix, 'ytitle', 'MMS'+str(this_probe)+' EDI drift velocity')
                    options('mms'+str(this_probe)+'_'+instrument+'_vdrift_gse_'+this_dr+'_'+this_lvl+suffix, 'legend_names', ['Vx GSE', 'Vy GSE', 'Vz GSE'])
                if 'mms'+str(this_probe)+'_'+instrument+'_vdrift_gsm_'+this_dr+'_'+this_lvl+suffix in tvars:
                    options('mms'+str(this_probe)+'_'+instrument+'_vdrift_gsm_'+this_dr+'_'+this_lvl+suffix, 'ytitle', 'MMS'+str(this_probe)+' EDI drift velocity')
                    options('mms'+str(this_probe)+'_'+instrument+'_vdrift_gsm_'+this_dr+'_'+this_lvl+suffix, 'legend_names', ['Vx GSM', 'Vy GSM', 'Vz GSM'])
                if 'mms'+str(this_probe)+'_'+instrument+'_e_dsl_'+this_dr+'_'+this_lvl+suffix in tvars:
                    options('mms'+str(this_probe)+'_'+instrument+'_e_dsl_'+this_dr+'_'+this_lvl+suffix, 'ytitle', 'MMS'+str(this_probe)+' EDI e-field')
                    options('mms'+str(this_probe)+'_'+instrument+'_e_dsl_'+this_dr+'_'+this_lvl+suffix, 'legend_names', ['Ex DSL', 'Ey DSL', 'Ez DSL'])
                if 'mms'+str(this_probe)+'_'+instrument+'_e_gse_'+this_dr+'_'+this_lvl+suffix in tvars:
                    options('mms'+str(this_probe)+'_'+instrument+'_e_gse_'+this_dr+'_'+this_lvl+suffix, 'ytitle', 'MMS'+str(this_probe)+' EDI e-field')
                    options('mms'+str(this_probe)+'_'+instrument+'_e_gse_'+this_dr+'_'+this_lvl+suffix, 'legend_names', ['Ex GSE', 'Ey GSE', 'Ez GSE'])
                if 'mms'+str(this_probe)+'_'+instrument+'_e_gsm_'+this_dr+'_'+this_lvl+suffix in tvars:
                    options('mms'+str(this_probe)+'_'+instrument+'_e_gsm_'+this_dr+'_'+this_lvl+suffix, 'ytitle', 'MMS'+str(this_probe)+' EDI e-field')
                    options('mms'+str(this_probe)+'_'+instrument+'_e_gsm_'+this_dr+'_'+this_lvl+suffix, 'legend_names', ['Ex GSM', 'Ey GSM', 'Ez GSM'])


import pyspedas
from pytplot import tplot, tplot_names
from pyspedas.themis.spacecraft.fields.fit import cal_fit

t = ['2008-03-15', '2008-03-16']
pyspedas.themis.fit(trange=t, probe='a', level='l1', get_support_data=True,
                    varnames=['tha_fit', 'tha_fit_code', 'tha_fit_npts'], time_clip=True)
cal_fit(probe='a')

tplot_names()
tplot(['tha_fit_bfit', 'tha_fit_efit'])
#tplot(['tha_fgs', 'tha_fgs_sigma', 'tha_fit_bfit', 'tha_fit_efit',
#       'tha_efs', 'tha_efs_sigma', 'tha_efs_0', 'tha_efs_dot0'])



from .gmag_isee_induction import gmag_isee_induction

def gmag_stel_induction(
    trange=['2018-10-18/00:00:00','2018-10-18/02:00:00'],
    suffix='',
    site='all',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    uname=None,
    passwd=None,
    time_clip=False,
    ror=True,
    frequency_dependent=False
):

    return gmag_isee_induction(
                                trange=trange,
                                suffix=suffix,
                                site=site,
                                get_support_data=get_support_data,
                                varformat=varformat,
                                varnames=varnames,
                                downloadonly=downloadonly,
                                notplot=notplot,
                                no_update=no_update,
                                uname=uname,
                                passwd=passwd,
                                time_clip=time_clip,
                                ror=ror,
                                frequency_dependent=frequency_dependent
                                )



"""
Subtracts the average (mean) or the median from the data.

Notes
-----
Similar to tsub_average.pro in IDL SPEDAS.

"""
import logging
import pyspedas
import pytplot
import numpy


def subtract_average(names, new_names=None, suffix=None, overwrite=None,
                     median=None):
    """
    Subtracts the average or the median from data.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    new_names: str/list of str, optional
        List of new_names for pytplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-d'.
    overwrite: bool, optional
        If set, then pytplot variables are replaced.
    median: float, optional
        If it is 0 or not set, then it computes the mean.
        Otherwise, it computes the median.

    Returns
    -------
    None.

    """
    old_names = pyspedas.tnames(names)

    if len(old_names) < 1:
        logging.error('Subtract Average error: No pytplot names were provided.')
        return

    if suffix is None:
        if median:
            suffix = '-m'
        else:
            suffix = '-d'

    if overwrite is not None:
        n_names = old_names
    elif new_names is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = new_names

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]

    old_names = pyspedas.tnames(names)

    for old_idx, old in enumerate(old_names):
        new = n_names[old_idx]

        if new != old:
            pyspedas.tcopy(old, new)

        data = pytplot.data_quants[new].values
        dim = data.shape
        if median:
            if len(dim) == 1:
                data -= numpy.median(data, axis=0)
            else:
                for i in range(dim[1]):
                    data[:, i] -= numpy.median(data[:, i], axis=0)
            ptype = 'Median'
        else:
            if len(dim) == 1:
                data -= numpy.mean(data, axis=0)
            else:
                for i in range(dim[1]):
                    data[:, i] -= numpy.mean(data[:, i], axis=0)
            ptype = 'Mean'

        pytplot.data_quants[new].values = data

        logging.info('Subtract ' + ptype + ' was applied to: ' + new)



import logging
from pytplot import get_data, store_data


def mms_fpi_split_tensor(tensor_variable):
    """
    Splits FPI tensor variables (pressure, temperature) into their components

    Input
    -----------
        tensor_variable: str
            Tplot variable containing the tensor data

    Returns
    ---------
        List of variables created.
    """

    data = get_data(tensor_variable, dt=True)

    if data is None:
        logging.error('Problem returning data from the variable: ' + tensor_variable)
        return

    saved = store_data(tensor_variable + '_xx', data={'x': data.times, 'y': data.y[:, 0, 0]})
    if not saved:
        logging.warning('Problem saving xx component')
    saved = store_data(tensor_variable + '_xy', data={'x': data.times, 'y': data.y[:, 0, 1]})
    if not saved:
        logging.warning('Problem saving xy component')
    saved = store_data(tensor_variable + '_xz', data={'x': data.times, 'y': data.y[:, 0, 2]})
    if not saved:
        logging.warning('Problem saving xz component')
    saved = store_data(tensor_variable + '_yx', data={'x': data.times, 'y': data.y[:, 1, 0]})
    if not saved:
        logging.warning('Problem saving yx component')
    saved = store_data(tensor_variable + '_yy', data={'x': data.times, 'y': data.y[:, 1, 1]})
    if not saved:
        logging.warning('Problem saving yy component')
    saved = store_data(tensor_variable + '_yz', data={'x': data.times, 'y': data.y[:, 1, 2]})
    if not saved:
        logging.warning('Problem saving yz component')
    saved = store_data(tensor_variable + '_zx', data={'x': data.times, 'y': data.y[:, 2, 0]})
    if not saved:
        logging.warning('Problem saving zx component')
    saved = store_data(tensor_variable + '_zy', data={'x': data.times, 'y': data.y[:, 2, 1]})
    if not saved:
        logging.warning('Problem saving zy component')
    saved = store_data(tensor_variable + '_zz', data={'x': data.times, 'y': data.y[:, 2, 2]})
    if not saved:
        logging.warning('Problem saving zz component')

    components = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz']
    return [tensor_variable + '_' + component for component in components]




import numpy as np
import logging
import re
from pyspedas.mms.mms_load_data import mms_load_data
from pyspedas.mms.hpca.mms_hpca_set_metadata import mms_hpca_set_metadata
from pyspedas.mms.hpca.mms_get_hpca_info import mms_get_hpca_info
from pyspedas.mms.hpca.mms_hpca_energies import mms_hpca_energies
from pyspedas.mms.print_vars import print_vars
from pyspedas.mms.mms_config import CONFIG

from pytplot import get_data, store_data

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

@print_vars
def mms_load_hpca(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy', 
    level='l2', datatype='moments', get_support_data=None, time_clip=False, no_update=False,
    varformat=None, varnames=[], suffix='', center_measurement=False, available=False, notplot=False, 
    latest_version=False, major_version=False, min_version=None, cdf_version=None, spdf=False,
    always_prompt=False):
    """
    This function loads HPCA data into tplot variables
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate : str or list of str
            instrument data rates for HPCA include 'brst', 'srvy'. The
            default is 'srvy'.

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'

        datatype : str or list of str
            Valid datatypes for HPCA are 'moments' and 'ion'; the default is 'moments'

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        center_measurement: bool
            If True, the CDF epoch variables are time-shifted to the middle
            of the accumulation interval by their DELTA_PLUS_VAR and
            DELTA_MINUS_VAR variable attributes

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multi-dimensional data products)

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested paramters

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')

        min_version: str
            Specify a minimum CDF version # to load

        latest_version: bool
            Only grab the latest CDF version in the requested time interval

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidently save an incorrect password, or if your SDC password has changed

        spdf: bool
            If True, download the data from the SPDF instead of the SDC
            
    Returns:
        List of tplot variables created.

    """

    if level.lower() != 'l2':
        if varformat is None:
            if level.lower() != 'l1a':
                if datatype.lower() == 'ion':
                    varformat = '*'
                elif datatype.lower() == 'combined':
                    varformat = '*'
                elif datatype.lower() == 'rf_corr':
                    varformat = '*_RF_corrected'
                elif datatype.lower() == 'count_rate':
                    varformat = '*_count_rate'
                elif datatype.lower() == 'flux':
                    varformat = '*_flux'
                elif datatype.lower() == 'vel_dist':
                    varformat = '*_vel_dist_fn'
                elif datatype.lower() == 'bkgd_corr':
                    varformat = '*_bkgd_corrected'
                elif datatype.lower() == 'moments':
                    varformat = '*'
                else:
                    varformat = '*_RF_corrected'
                if varformat != None and varformat != '*':
                    datatype = 'ion'
                if level.lower() == 'sitl':
                    varformat = '*'
    else:
        if get_support_data is None:
            get_support_data = True

        if isinstance(datatype, str):
            if datatype not in ['ion', 'moments']:
                logging.warning("Unknown datatype: " + datatype + "for L2 HPCA data; expected 'ion' or 'moments', loading 'ion'")
                datatype = 'ion'

    tvars = mms_load_data(trange=trange, notplot=notplot, probe=probe, data_rate=data_rate, level=level, instrument='hpca',
            datatype=datatype, varformat=varformat, varnames=varnames, suffix=suffix, get_support_data=get_support_data,
            time_clip=time_clip, no_update=no_update, center_measurement=center_measurement, available=available, 
            latest_version=latest_version, major_version=major_version, min_version=min_version, cdf_version=cdf_version,
            spdf=spdf, always_prompt=always_prompt)
    
    if tvars is None or available or notplot or CONFIG['download_only']:
        return tvars

    mms_hpca_set_metadata(probe=probe, suffix=suffix)

    if not isinstance(datatype, list):
        datatype = [datatype]
    if not isinstance(probe, list):
        probe = [probe]

    # Replace supplementary fields in 3D distribution variables with actual
    # values from supplementary tplot variables (theta).
    for dtype in datatype:
        if dtype == 'moments':
            continue

        for prb in probe:
            prb = str(prb)
            for tvar in tvars:
                df_var = re.search('^mms'+prb+'_hpca_[^_]+plus(_(phase_space_density|count_rate|flux)| ?)'+suffix+'$', tvar)
                if df_var:
                    df_data = get_data(tvar)
                    df_metadata = get_data(tvar, metadata=True)
                    theta_data = get_data('mms' + prb + '_hpca_centroid_elevation_angle' + suffix)

                    if theta_data is None:
                        info = mms_get_hpca_info()
                        theta = info['elevation']
                    else:
                        theta = theta_data

                    store_data(tvar, data={'x': df_data.times, 'y': df_data.y, 'v1': theta, 'v2': df_data.v2}, attr_dict=df_metadata)

                    # check if energy table contains all 0s
                    zerocheck = np.argwhere(df_data.v2 == 0.0)
                    if len(zerocheck) == 63:
                        #  energy table is all 0s, using hard coded table
                        energy_table = mms_hpca_energies()
                        logging.warning('Found energy table with all 0s: ' + tvar + '; using hard-coded energy table instead')
                        store_data(tvar, data={'x': df_data.times, 'y': df_data.y, 'v1': theta, 'v2': energy_table}, attr_dict=df_metadata)

    return tvars



import os
import requests
import logging
import warnings
import pkg_resources
import numpy as np
from pytplot import cdf_to_tplot
from ..analysis.time_clip import time_clip as tclip
from pyspedas import time_double, time_string
from dateutil.parser import parse
from datetime import timedelta, datetime
from shutil import copyfileobj, copy
from tempfile import NamedTemporaryFile
from .mms_config import CONFIG
from .mms_get_local_files import mms_get_local_files
from .mms_files_in_interval import mms_files_in_interval
from .mms_login_lasp import mms_login_lasp
from .mms_file_filter import mms_file_filter
from .mms_load_data_spdf import mms_load_data_spdf


def mms_load_data(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy', level='l2', 
    instrument='fgm', datatype='', varformat=None, prefix='', suffix='', get_support_data=False, time_clip=False, 
    no_update=False, center_measurement=False, available=False, notplot=False, latest_version=False, 
    major_version=False, min_version=None, cdf_version=None, spdf=False, always_prompt=False, varnames=[]):
    """
    This function loads MMS data into pyTplot variables

    This function is not meant to be called directly. Please see the individual load routines for documentation and use. 

    """

    if not isinstance(probe, list): probe = [probe]
    if not isinstance(data_rate, list): data_rate = [data_rate]
    if not isinstance(level, list): level = [level]
    if not isinstance(datatype, list): datatype = [datatype]
    
    probe = [str(p) for p in probe]

    # allows the user to pass in trange as list of datetime objects
    if type(trange[0]) == datetime and type(trange[1]) == datetime:
        trange = [time_string(trange[0].timestamp()), time_string(trange[1].timestamp())]

    # allows the user to pass in trange as a list of floats (unix times)
    if isinstance(trange[0], float):
        trange[0] = time_string(trange[0])
    if isinstance(trange[1], float):
        trange[1] = time_string(trange[1])
        

    download_only = CONFIG['download_only']

    no_download = False
    if no_update or CONFIG['no_download']: no_download = True

    if spdf:
        return mms_load_data_spdf(trange=trange, probe=probe, data_rate=data_rate, level=level, 
                                  instrument=instrument, datatype=datatype, varformat=varformat, 
                                  suffix=suffix, get_support_data=get_support_data, time_clip=time_clip, 
                                  no_update=no_update, center_measurement=center_measurement, notplot=notplot, 
                                  latest_version=latest_version, major_version=major_version, 
                                  min_version=min_version, cdf_version=cdf_version, varnames=varnames)

    headers = {}
    try:
        release_version = pkg_resources.get_distribution("pyspedas").version
    except pkg_resources.DistributionNotFound:
        release_version = 'bleeding edge'
    headers['User-Agent'] = 'pySPEDAS ' + release_version

    user = None
    if not no_download:
        sdc_session, user = mms_login_lasp(always_prompt=always_prompt, headers=headers)

    out_files = []
    available_files = []

    for prb in probe:
        for drate in data_rate:
            start_date = parse(trange[0]).strftime('%Y-%m-%d') # need to request full day, then parse out later
            end_date = parse(time_string(time_double(trange[1])-0.1)).strftime('%Y-%m-%d-%H-%M-%S') # -1 second to avoid getting data for the next day
            # kludge to fix issue for burst mode data in files from the previous day
            if drate == 'brst':
                sec_from_start_of_day = time_double(trange[0])-time_double(start_date)

                # check if we're within 10 minutes of the start of the day
                # and if so, grab 10 minutes of data from the end of the
                # previous day
                if sec_from_start_of_day <= 600.0:
                    start_date = time_string(time_double(start_date)-600.0, fmt='%Y-%m-%d-%H-%M-%S')

            for lvl in level:
                for dtype in datatype:
                    if user is None:
                        url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/file_info/science?start_date=' + start_date + '&end_date=' + end_date + '&sc_id=mms' + prb + '&instrument_id=' + instrument + '&data_rate_mode=' + drate + '&data_level=' + lvl
                    else:
                        url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/file_info/science?start_date=' + start_date + '&end_date=' + end_date + '&sc_id=mms' + prb + '&instrument_id=' + instrument + '&data_rate_mode=' + drate + '&data_level=' + lvl
                    
                    if dtype != '':
                        url = url + '&descriptor=' + dtype

                    if CONFIG['debug_mode']: logging.info('Fetching: ' + url)

                    if no_download == False:
                        # query list of available files
                        try:
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore", category=ResourceWarning)
                                http_json = sdc_session.get(url, verify=True, headers=headers).json()

                            if CONFIG['debug_mode']: logging.info('Filtering the results down to your trange')

                            files_in_interval = mms_files_in_interval(http_json['files'], trange)

                            if available:
                                for file in files_in_interval:
                                    logging.info(file['file_name'] + ' (' + str(np.round(file['file_size']/(1024.*1024), decimals=1)) + ' MB)')
                                    available_files.append(file['file_name'])
                                continue

                            for file in files_in_interval:
                                file_date = parse(file['timetag'])
                                if dtype == '':
                                    out_dir = os.sep.join([CONFIG['local_data_dir'], 'mms'+prb, instrument, drate, lvl, file_date.strftime('%Y'), file_date.strftime('%m')])
                                else:
                                    out_dir = os.sep.join([CONFIG['local_data_dir'], 'mms'+prb, instrument, drate, lvl, dtype, file_date.strftime('%Y'), file_date.strftime('%m')])

                                if drate.lower() == 'brst':
                                    out_dir = os.sep.join([out_dir, file_date.strftime('%d')])

                                out_file = os.sep.join([out_dir, file['file_name']])

                                if CONFIG['debug_mode']: logging.info('File: ' + file['file_name'] + ' / ' + file['timetag'])

                                if os.path.exists(out_file) and str(os.stat(out_file).st_size) == str(file['file_size']):
                                    if not download_only: logging.info('Loading ' + out_file)
                                    out_files.append(out_file)
                                    continue

                                if user is None:
                                    download_url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/download/science?file=' + file['file_name']
                                else:
                                    download_url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/download/science?file=' + file['file_name']

                                logging.info('Downloading ' + file['file_name'] + ' to ' + out_dir)

                                with warnings.catch_warnings():
                                    warnings.simplefilter("ignore", category=ResourceWarning)
                                    fsrc = sdc_session.get(download_url, stream=True, verify=True, headers=headers)
                                ftmp = NamedTemporaryFile(delete=False)

                                with open(ftmp.name, 'wb') as f:
                                    copyfileobj(fsrc.raw, f)

                                if not os.path.exists(out_dir):
                                    os.makedirs(out_dir)

                                # if the download was successful, copy to data directory
                                copy(ftmp.name, out_file)
                                out_files.append(out_file)
                                fsrc.close()
                                ftmp.close()
                        except requests.exceptions.ConnectionError:
                            # No/bad internet connection; try loading the files locally
                            logging.error('No internet connection!')

                    
                    if out_files == []:
                        if not download_only: logging.info('Searching for local files...')
                        out_files = mms_get_local_files(prb, instrument, drate, lvl, dtype, trange)

                        if out_files == [] and CONFIG['mirror_data_dir'] != None:
                            # check for network mirror; note: network mirrors are assumed to be read-only
                            # and we always copy the files from the mirror to the local data directory
                            # before trying to load into tplot variables 
                            logging.info('No local files found; checking network mirror...')
                            out_files = mms_get_local_files(prb, instrument, drate, lvl, dtype, trange, mirror=True)

    if not no_download:
        sdc_session.close()

    if available:
        return available_files

    if not download_only:
        out_files = sorted(out_files)

        filtered_out_files = mms_file_filter(out_files, latest_version=latest_version, major_version=major_version, min_version=min_version, version=cdf_version)
        if filtered_out_files == []:
            logging.info('No matching CDF versions found.')
            return

        new_variables = cdf_to_tplot(filtered_out_files, varformat=varformat, varnames=varnames, get_support_data=get_support_data, prefix=prefix, suffix=suffix, center_measurement=center_measurement, notplot=notplot)

        if notplot:
            return new_variables

        if new_variables == []:
            logging.warning('No data loaded.')
            return

        if time_clip:
            for new_var in new_variables:
                tclip(new_var, trange[0], trange[1], suffix='')

        return new_variables
    else:
        return out_files








import cdflib
import logging
import numpy as np


from copy import deepcopy
from scipy.spatial import KDTree
from pyspedas import tnames
from pyspedas.utilities.time_double import time_double
from pyspedas.utilities.time_string import time_string
from pytplot import get_data
from scipy import interpolate

from astropy.coordinates import spherical_to_cartesian, cartesian_to_spherical

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)
def erg_hep_get_dist(tname,
                      index=None,
                      units='flux',
                      level='l2',
                      species='e',
                      time_only=False,
                      single_time=None,
                      trange=None,
                      new_effic=False,
                      w_sct015=False,
                      exclude_azms=False):
    if len(tnames(tname)) > 0:
        input_name = tnames(tname)[0]
    else:
        print(f'Variable: {tname} not found!')
        return 0
    level = level.lower()
    
    #  ;; only erg_hep_l2_3dflux_FEDU is acceptable for this routine
    if ('erg_hep_l2_FEDU_L' not in input_name) \
    and ('erg_hep_l2_FEDU_H' not in input_name) \
    and ('erg_hep_l2_rawcnt_H' not in input_name) \
    and ('erg_hep_l2_rawcnt_L' not in input_name):
        print(f'Variable {input_name}  is not acceptable. This routine can take only erg_hep_l2_(FEDU/rawcnt)_? as the argument.')
        return 0
    """
    ;; Extract some information from a tplot variable name
    ;; e.g., erg_lepi_l2_3dflux_FPDU
    """
    vn_info = input_name.split('_')
    instrument = vn_info[1]  #  ;;hep
    level = vn_info[2]       #  ;;l2
    dtype = vn_info[3]       #  ;;FEDU or rawcnt
    suf = vn_info[4]        #  ;; L or H 
    arrnm = vn_info[4]
    if species is None:
        if instrument == 'hep':
            species = 'e'
        else:
            print(f'RROR: given an invalid tplot variable: {input_name}')
    # ;; Reform a data array so that it is grouped for each spin
    data_in = get_data(input_name)
    
    if data_in is None:
        print('Problem extracting the mepe 3dflux data.')
        return 0
    data_in_metadata = get_data(input_name, metadata=True)
    t_fedu = deepcopy(data_in[0])
    fedu = deepcopy(data_in[1])
    vn_angsga = '_'.join(vn_info[0:3] +['FEDU', suf, 'Angle_sga'])
    if get_data(vn_angsga) is not None:
        angsga = deepcopy(get_data(vn_angsga)[1])
    else:
        print(f'Cannot find variable {vn_angsga}, which is essential for this routine to work.')
        return 0

    vn_sctno = '_'.join(vn_info[0:3] +['sctno', suf])
    if len(tnames(vn_sctno)) < 1:
        print(f'Cannot find variable {vn_sctno}, which is essential for this routine to work.')
        return 0
    else:
        time_scno, scno = get_data(tnames(vn_sctno)[0])

    id_scno_0=np.argwhere(scno == 0)
    sc0num = len(id_scno_0)
    if sc0num < 5:
        print(f'Only data for less than 5 spins are loaded for HEP_{suf}!!')
        return 0
    
    # ;; integration time for each spin sector
    sctdt = time_scno[1:] - time_scno[:-1]
    sctintgt=np.insert(sctdt, sctdt.shape[0],sctdt[-1])

    id_scno_0_list = id_scno_0[:,0].tolist()
    sc0_t = time_scno[id_scno_0_list]  # ;; times of spin sector #0, namely the start of each spin
    dt_array = sc0_t[1:] - sc0_t[:-1]
    sc0_dt = np.insert(dt_array, dt_array.shape[0], dt_array[-1])

    #  processing for 'id_t_fedu = value_locate()' in IDL.
    sc0_t_repeat = np.repeat(np.array([sc0_t]).T, 2, 1)
    t_fedu_repeat = np.repeat(np.array([t_fedu]).T, 2, 1)
    tree = KDTree(sc0_t_repeat)
    _, id_t_fedu_nn = tree.query(t_fedu_repeat)
    id_t_fedu = np.where((t_fedu - sc0_t[id_t_fedu_nn.tolist()]) < 0., id_t_fedu_nn-1, id_t_fedu_nn)
    #  processing for 'id_t_fedu = value_locate()' in IDL.

    #  ;; Genrate angarr by picking up angle values at each spin start
    angarr = angsga[id_scno_0_list, :, :]

    #  ;; fedu_arr:[ time, spin sct, energy, azm ]
    #   ;; by default
    nene = 16
    nazm = 15
    nsct = 16
    #   ;; by default

    if suf == 'H':  # ;; Lv2 HEP-H flux array has 11 elements for energy bin currently.
        nene = 11
    fedu_arr = np.full(shape=(sc0num, nsct, nene, nazm ), fill_value=np.nan)  # ;; padded with NaN
    intgt = np.full(shape=(sc0num, nsct), fill_value=np.nan)  # ;; padded with NaN
    
    for i in range(sc0num):
        id_array = np.argwhere( (id_t_fedu == i) & (scno >= 0) & (scno <= 15))
        if len(id_array) < 2:
            continue
        id_list = id_array[:, 0].tolist()
        datarr = fedu[id_list, :, :]  # ;; usually [16(time), 16(energy), 15(azm)]
        tarr = t_fedu[id_list]
        sctarr = scno[id_list]
        sctarr_list = sctarr.tolist()
        nsct = len(sctarr_list)
        inttarr = sctintgt[id_list]

        fedu_arr[i, sctarr_list, :, :] = datarr.reshape(1, nsct, nene, nazm)
        intgt[i, sctarr_list] = inttarr.reshape(1, nsct)

    p_structure = {
    'x':sc0_t,  # ;; Put the start time of each spin
    'y':fedu_arr  # ;; [ time, 16(sct), 16(energu), 15(azm) ]
    }

    if p_structure['y'].ndim != 4:
        print(f'Variable: {input_name} contains wrong number of elements!')
        return 0

    # ;; Return time labels corresponding the middle time of each spin
    if time_only:
        return p_structure['x']
    
    if single_time is not None:
        single_time_double = time_double(single_time)
        if (p_structure['x'][0] <= single_time_double)\
                and (single_time_double <= p_structure['x'][-1]):
            nearest_time = interpolate.interp1d(p_structure['x'], p_structure['x'],
                                                kind="nearest")(single_time_double)
        elif (single_time_double < p_structure['x'][0])\
                or (p_structure['x'][-1] < single_time_double):
            nearest_time = interpolate.interp1d(p_structure['x'], p_structure['x'],
                                                kind="nearest", fill_value='extrapolate')(single_time_double)
        index = np.where(p_structure['x'] == nearest_time)[0]
        n_times = index.size
    else:
        # index supersedes time range
        if index is None:
            if trange is not None:
                trange_double_array = np.array(time_double(trange))
                trange_minmax = np.array([trange_double_array.min(),
                                          trange_double_array.max()])
                index = np.where((p_structure['x'] >= trange_minmax[0])
                                 & (p_structure['x'] <= trange_minmax[1]))[0]
                n_times = index.size
                if n_times == 0:
                    print('No data in time range: '
                          + ' '.join(time_string(trange_minmax)))
                    return 0
            else:
                n_times = p_structure['x'].size
                index = np.arange(n_times)
        else:
            n_times = np.array(tuple([index])).size
    """
    ;; --------------------------------------------------------------
    ;; HEP data arr: [9550(time), 16(spin phase), 16(energy), 15(azimuth ch )]
    ;; Dimensions
    """
    # ;; Dimensions
    # ;; to [ energy, spin phase, azimuth ch(elevation) ]
    dim_array = np.array(p_structure['y'].shape[1:])[[1, 0, 2]]
    if species.lower() == 'e':
        mass = 5.68566e-06
        charge = -1.
        data_name = f'HEP-{suf} Electron 3dflux'
        if dtype == 'rawcnt':
            data_name = f'HEP-{suf} Electron raw count/sample'
        integ_time = 7.99 / 16  # ;; currently hard-coded, but practically not used.
    else:
        print('given species is not supported by this routine.')
        return 0
    #  ;; basic template structure compatible with other routines
    dist = {
        'project_name': 'ERG',
        'spacecraft': 1,  # always 1 as a dummy value
        'data_name': data_name,
        'units_name': 'flux',  # ; HEP data in [/keV-s-sr-cm2] should be converted to [/eV-s-sr-cm2]
        'units_procedure': 'erg_convert_flux_units',
        'species': species,
        'valid': 1,
        'charge': charge,
        'mass': mass,
    }
    # other keys (members) will added in below processing, sequentially.
    """
    ;; Then, fill in arrays in the data structure
    ;;   dim[ nenergy, nspinph(azimuth), nanode(elevation), ntime]
    """
    dist['time'] = p_structure['x'][tuple([index])]  # ;; the start time of spin
    dist['end_time'] = dist['time'] + sc0_dt[tuple([index])]  # ;; the end time of spin
    """
    ;; Shuffle the original data array [time,spin phase,energy,apd] to
    ;; be energy-azimuth-elevation-time.
    ;; The factor 1d-3 is to convert [/keV-s-sr-cm2] (default unit of
    ;; HEP Lv2 flux data) to [/eV-s-sr-cm2] 
    """
    if 'cnt' not in dtype:
        dist['data'] = p_structure['y'][tuple([index])].transpose([2, 1, 3, 0]) * 1.e-03
    else:
        dist['data'] = p_structure['y'][tuple([index])].transpose([2, 1, 3, 0])  # ;; for count/sample, count/sec
    
    dist['bins'] = np.ones(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), dtype='int8')

    """
    ;; Exclude spin phases #0 and #15 currently from spectrum
    ;; calculations
    """
    if not w_sct015:
        dist['bins'][:, 0, :, :] = 0
        dist['bins'][:, 15, :, :] = 0
    
    #  ;; Exclude invalid azimuth channels
    if exclude_azms:
        invalid_azms = [0, 4, 5, 9, 10, 11, 14]
        dist['bins'][:, :, invalid_azms, :] = 0

    #  ;; Apply the empiricallly-derived efficiency (only for cnt/cntrate/dtcntrate)
    if (new_effic) and ('cnt' in dtype):
        """
        ;; efficiency for each azim. ch. based on the inter-ch. calibration
        ;; Only valid for 2017-06-21 through 2019-02-07 
        """
        effic = [
          0.171, 0.460, 1.013, 0.411, 0.158, 
          0.162, 0.450, 1.000, 0.399, 0.158, 
          0.120, 0.170, 0.629, 0.346, 0.132
        ]
        for i in range(15):
            dist['data'][:, :, i, :] /= effic[i]
        print(' new_effic has been applied!')

    file_name_raw= get_data(input_name, metadata=True)['CDF']['FILENAME']
    if isinstance(file_name_raw, str):
        cdf_path = file_name_raw
    elif isinstance(file_name_raw, list):
        cdf_path = file_name_raw[-1]

    cdf_file = cdflib.CDF(cdf_path)

    #  ;; Energy ch
    """
    ;;; Extract necessary information from the Lv2 data CDF file
    """
    enearr = cdf_file.varget('FEDU_'+suf+'_Energy')  # ;; [2(min,max), 16 (ene ch) ]
    # ;; enearr[0, 0] = 30                 #   ;; the effective lowest energy limit is assumed to be 30 keV

    """
    ;; Calculate averages, which may be replaced with a more
    ;; appropriate averaging method for representative energies. 
    ;; Currently the geometric average is adopted. 
    ;; 1e3 is to convert [keV] (default of HEP Lv2 flux data) to [eV]
    """
    e0_array = np.sqrt(enearr[0, :] * enearr[1, :]) * 1.e+03
    energy_reform = np.reshape(e0_array, [dim_array[0], 1, 1, 1])
    energy_rebin1 = np.repeat(energy_reform, dim_array[2],
                             axis=2)  # repeated across apd(elevation)
    energy_rebin2 = np.repeat(energy_rebin1, dim_array[1],
                              axis=1)  # repeated across spin phase(azimuth)
    dist['energy'] = np.repeat(energy_rebin2, n_times,
                              axis=3)  # repeated across n_time

    # denergy member
    
    de_array = (enearr[1, :] - enearr[0, :]) * 1.e+03

    de_reform = np.reshape(de_array, [dim_array[0], 1, 1, 1])
    de_rebin1 = np.repeat(de_reform, dim_array[2],
                         axis=2)  # repeated across apd(elevation)
    de_rebin2 = np.repeat(de_rebin1, dim_array[1],
                         axis=1)  # repeated across spin phase(azimuth)
    dist['denergy'] = np.repeat(de_rebin2, n_times,
                         axis=3)  # repeated across n_times


    dist['n_energy'] = dim_array[0]
    dist['n_bins'] = dim_array[1] * dim_array[2]  #   # thetas * # phis
    """
    ;; Array elements containing NaN are excluded from the further
    ;; calculations, by setting bins to be zero.
    """

    # ;; converted to the flux dirs
    unitvec = np.ones_like(angarr[:, 0, :])

    deg_to_rad = np.pi / 180.
    ex_array, ey_array, ez_array = spherical_to_cartesian(
                                     unitvec,
                                     angarr[:, 0, :] * deg_to_rad,
                                     angarr[:, 1, :] * deg_to_rad
                                     )
    r_array, theta_array, phi_array = cartesian_to_spherical(
                                        -ex_array,
                                        -ey_array,
                                        -ez_array)
    theta_array = theta_array.value / deg_to_rad
    phi_array = phi_array.value / deg_to_rad

    #  ;; --> [(time), 2, 15]
    angarr[:,0,:] = theta_array
    angarr[:,1,:] = phi_array
    #  ;; --> [(time), 2, 15]

    spinper=sc0_dt[tuple([index])]
    if len(spinper) == 1:
        spinper = spinper[0]

    rel_sct_time = np.zeros_like(intgt[tuple(index), :])
    #  ;; Elapsed time from spin start through the center of each spin sector
    for i in range(16):
        if i == 0:
            rel_sct_time[:, i] = 0. + intgt[tuple(index), i]/2
        elif i == 1:
            rel_sct_time[:, i] = intgt[tuple(index), 0] + intgt[tuple(index), i]/2
        else:
            rel_sct_time[:, i] = np.nansum(intgt[tuple(index), 0:(i)], axis=1) + intgt[tuple(index), i]/2



    phissi = angarr[:, 1, :] - (90. + 21.6)  #;; [ (time), (azm)]
    spinper_rebin = np.repeat(spinper.reshape(n_times, 1), 16, axis=1)
    spinph_ofst = rel_sct_time / spinper_rebin * 360.
    phi0_1_reform = np.reshape(phissi.T, [1, 1, dim_array[2], n_times])
    phi0_1_rebin1 = np.repeat(phi0_1_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)
    phi0_1 = np.repeat(phi0_1_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    phi0_2_reform = np.reshape(spinph_ofst, [n_times, 1, dim_array[1], 1])
    phi0_2_rebin1 = np.repeat(phi0_2_reform, dim_array[2],
                             axis=3)  # repeated across apd(elevation)
    phi0_2 = np.repeat(phi0_2_rebin1, dim_array[0],
                             axis=1)  # repeated across energy

    phi0 = phi0_1 + phi0_2.transpose(1, 2, 3, 0)

    """
    ;;  phi angle for the start of each spin phase
    ;;    + offset angle 
    """
    dist['phi'] = np.fmod((phi0 + 360.), 360.)
    dist['dphi'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), fill_value=22.5)  # ;; 22.5 deg as a constant
    del phi0  # ;; Clean huge arrays
    dist['n_phi'] = dim_array[1]
    #  ;; elevation angle
    elev = angarr[:, 0, :]  # ;; [ (time), (Az.ch)]
    elev_reform = np.reshape(elev.T, [1, 1, dim_array[2], n_times])
    elev_rebin1 = np.repeat(elev_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)
    dist['theta'] = np.repeat(elev_rebin1, dim_array[0],
                             axis=0)  # repeated across energy

    dist['dtheta'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), fill_value=12.0)

    dist['n_theta'] = dim_array[2]
    
    return dist



import numpy as np

from .erg_convert_flux_units import erg_convert_flux_units


def erg_pgs_clean_data(data_in,
                       units='flux',
                       relativistic=False,
                       for_moments=False,
                       magf=np.array([0., 0., 0.]),
                       muconv=False
                       ):

    converted_data = erg_convert_flux_units(input_dist=data_in,
                                            units=units,
                                            relativistic=relativistic)

    dims = np.array(converted_data['data'].shape)
    angdims = converted_data['data'][0, :].size

    output = {
        'dims': dims,
        'time': converted_data['time'],
        'end_time': converted_data['end_time'],
        'charge': converted_data['charge'],
        'mass': converted_data['mass'],
        'species': converted_data['species'],
        'magf': magf,
        'sc_pot': 0.,
        'scaling': np.ones(shape=(dims[0], angdims)),
        'units_name': data_in['units_name'],
        'psd': data_in['data'].reshape(dims[0], angdims),
        'data': converted_data['data'].reshape(dims[0], angdims),
        'bins': converted_data['bins'].reshape(dims[0], angdims),
        'energy': converted_data['energy'].reshape(dims[0], angdims),
        'denergy': converted_data['denergy'].reshape(dims[0], angdims),
        'phi': converted_data['phi'].reshape(dims[0], angdims),
        'dphi': converted_data['dphi'].reshape(dims[0], angdims),
        'theta': converted_data['theta'].reshape(dims[0], angdims),
        'dtheta': converted_data['dtheta'].reshape(dims[0], angdims),
    }

    # Exclude f_nan values from further calculations
    bins = output['bins'].astype(np.int8)
    output['bins'] = np.where((bins == 0)
                              | (np.isinf(output['data']) == True)
                              | (np.isnan(output['data']) == True)
                              | (np.isinf(output['energy']) == True)
                              | (np.isnan(output['energy']) == True),
                              0, 1)

    # Fill invalid values with zero for moment calculations
    if for_moments:
        output['data'] = np.where(output['bins'] == 0, 0., output['data'])
        output['psd'] = np.where(output['bins'] == 0, 0., output['psd'])
        output['energy'] = np.where(output['bins'] == 0, 1., output['energy'])
        output['denergy'] = np.where(output['bins'] == 0,
                                    1., output['denergy'])

    if 'orig_energy' in converted_data.keys():
        output['orig_energy'] = converted_data['orig_energy']

    return output
